#!/usr/bin/perl 
use strict;
use warnings;
use lib '/usr/local/lib/mpmasql/modules';

use Storable qw(dclone);
use EmpiricalDF;
use Carp;
use POSIX;
use List::Util qw(max);
use Scalar::Util qw(looks_like_number);
use Data::Dumper qw(Dumper);
use Term::ReadKey;

use CtUtilities qw(cartesian is_in_list fisher_yates_shuffle determine_class_by_upperbound);
use IndexedTable;
use MpmasQLConfigReader;
use MpmasQLVariableProcessing;
use MpmasQLFunctionLanguage;
use MasMapAgent;
use MasMapRegion;
use EnhancedGrid;
use SimpleObject;
use SimpleObjectRegistry;


##CT110826 max
use MpmasMILP;


my $milpcheck = "/usr/local/bin/MilpCheck";
use constant INFTY => 10**31;


#$Carp::Verbose  = 1;
##Version:
my $thisversion = "v1.00";
my $lastchange = "29-Sep-2013";
my $starttime = time;

my $longwelcome = <<WC;


####################################################################
####			mpmasdist $thisversion		####
####	Perl script to distribute endowments to MP-MAS agents	####
####								####
####	(c)	Christian Troost, Univ. Hohenheim, 2011-2013	####
####	Version: $thisversion		$lastchange		####
####################################################################
WC

my $shortwelcome = <<WC2;
mpmasdist $thisversion		
(c)	Christian Troost, Universität Hohenheim, 2011-2013
WC2

my $ininame;
my %setstrings;
while (@ARGV) {
	my $element = shift @ARGV;
	if  ($element =~ /^-v$/) {
		print $longwelcome;
		exit 0;
	}
	elsif (not $element =~ /^-/) {
		# inifile name
		$ininame = $element;
		print "\nControl file:\t\t". $ininame."\n";  
	}
	elsif ($element =~ /^--set/ ) {
		if (not $ARGV[0] =~ /^-/ ) {
		    my $setstring = shift @ARGV;
		    my ($var, $val) = split /=/,$setstring,2;
		    $var =~ s/^\s*//;
		    $val =~ s/^\s*//;
		    $var =~ s/\s*$//;
		    $val =~ s/\s*$//;

		    $setstrings{$var} = $val;
		}
	}
}

#Give error message if no input file was specified
print STDERR "You forgot to specify a control file!\n" and exit(1) unless $ininame;

######## Read Ini-File
my ($inihash, $setup, $tables, $discollect, $ppinfo,$varrank,  $maps, $stepsclude) = MpmasQLConfigReader::ReadDst($ininame, 0);
######## Connect to Db
my $dbh;
ReadMode(0);
if (defined  $inihash->{DBTYPE} and defined $inihash->{DB}) {
	if (not $inihash->{DBPWD}) {
		print "Password: ";
		ReadMode 2;	
		 $inihash->{DBPWD} = ReadLine (0);
		print "\n";
		ReadMode 1;
	}
	chomp( $inihash->{DBPWD});
	$dbh = DBI->connect("dbi:".$inihash->{DBTYPE}.":". $inihash->{DB}, $inihash->{DBUSER}, $inihash->{DBPWD}, { RaiseError => 1, AutoCommit => 1 }) or die $!;
	print "Successfully connected to " . $inihash->{DB} . " as ". $inihash->{DBUSER} . "\n"; 
	$dbh->do("SET NAMES utf8");
}


#interprete variables
MpmasQLFunctionLanguage->set_globals(0,$dbh,{},{},0);
for my $i (sort {$varrank->{$a} <=> $varrank->{$b} } keys %$setup) {
  if (exists $setstrings{$i}) {
	$setup->{$i} = MpmasQLFunctionLanguage->convert($setstrings{$i}, {}) ;
  }
  else{
	$setup->{$i} = MpmasQLFunctionLanguage->convert($setup->{$i}, {});
  }
  MpmasQLFunctionLanguage->set_globals( 0, $dbh, $setup,"",0);
}
#mkpaths
if (exists $stepsclude->{MKPATHS}) {
  for my $p ( MpmasQLFunctionLanguage->convert($stepsclude->{MKPATHS}, {})) {
	system("mkdir -p ". $p);
  }
}

#load tables
my $tableregistry= SimpleObjectRegistry->create();
for my $l (keys %{$tables}){
    $tableregistry->add_objects($tables->{$l}->transform($dbh));
}
MpmasQLFunctionLanguage->set_globals( 0, $dbh, $setup,$tableregistry);
#load maps
my $mapregistry = SimpleObjectRegistry->create();
for my $m (keys %$maps) {
	
	my $mapfile = MpmasQLFunctionLanguage->convert($maps->{$m},{});
	my $newmap = EnhancedGrid->create_from_ascii($mapfile);
	$newmap->set_id($m);
	$mapregistry->add_objects($newmap);
}
MpmasQLFunctionLanguage->set_mapregistry($mapregistry);

my @steplist;
if (exists $stepsclude->{INCLUDE}) {
    @steplist = MpmasQLFunctionLanguage->convert($stepsclude->{INCLUDE},{});
}
else {
      @steplist = sort {$a <=> $b} keys %$discollect;
}
if (exists $stepsclude->{EXCLUDE}) {
  my @exclude = MpmasQLFunctionLanguage->convert($stepsclude->{EXCLUDE},{});
  my %testhash = map { $_ => 1} @exclude;
  my @newlist = map { exists $testhash{$_} ? () : $_  } @steplist;
  @steplist = @newlist;
}
print "Steps to process: ".join (",", @steplist)."\n";



#declare 
my $total_agent_registry = SimpleObjectRegistry->create();
my $notassigned = SimpleObjectRegistry->create();
my $subarearegistry = SimpleObjectRegistry->create();




for my $d (@steplist) {
	
	my $disinfo = $discollect->{$d}; 
	my $seed;
	if (exists $disinfo->{SEED}) {
		$seed = MpmasQLFunctionLanguage->convert($disinfo->{SEED},{});
		srand($seed) if (defined $seed); 
	}
	print "STEP $d: ";
	print $disinfo->{TYPE};
	print " SEED: ". $seed if (defined $disinfo->{SEED});
	my  ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = localtime(time);
	$mon += 1;
	$year += 1900;
	print "\tTIME: $year $mon $mday , $hour:$min:$sec\n";
	if (exists $disinfo->{DESCRIPTION}) {
		my $desc = MpmasQLFunctionLanguage->convert($disinfo->{DESCRIPTION},{});
		print $desc."\n" if (defined $desc);
	}
	my $verbose = $disinfo->{VERBOSE}||0;
	my $constraints = divide_constraints($disinfo);
	
	
	my %clusters;
	my $cc = 0;
	my @combined = ([(0)]);
	my @clulist = ("grp");
	if (exists $disinfo->{LOOPS}->{lines} and @{$disinfo->{LOOPS}->{lines}} > 0) {
		undef @clulist;
		for my $i (0..$#{$disinfo->{LOOPS}->{lines} }) {
			my ($field, $content) = split /\s*#=\s*/, $disinfo->{LOOPS}->{lines}->[$i];
			my $key = MpmasQLFunctionLanguage->convert($field,{});
			push @clulist, $key;
			$clusters{$key} = [(MpmasQLFunctionLanguage->convert($content,{}) )];	
		}
		
		
		undef @combined;
		
		if (scalar @clulist == 1) {
			@combined = map { [ $_ ] }  @{$clusters{$clulist[0]} };
		}
		else {
		  my @temp = map { $clusters{$_} } @clulist;
		  foreach (@clulist) {
			@combined = cartesian (@temp );
		  }
		}
	}
	my $agentregistry = SimpleObjectRegistry->create(); #declared outside foreach, so it is visible in continue, make sure it is always cleared!
	
	CLULOOP: foreach my $clu (@combined) {
		my $locvarhash = {}; 
		foreach my $i (0..$#clulist) {
			
			$locvarhash->{$clulist[$i]} = $clu->[$i];
		}
		print "\e[K \r\tLOOP: @$clu -> \n" if (scalar @combined > 1);
		#map {print join("\t", $_, $locvarhash->{$_})."\n" } keys %$locvarhash;

		
		if (not defined $disinfo->{AGENTS}->{list} or $disinfo->{AGENTS}->{list} =~ /^\s*$/ or $disinfo->{AGENTS}->{list} eq "AGENTS" ) {
			$agentregistry = dclone($total_agent_registry);	
			$total_agent_registry->clear();
		}
		elsif ($disinfo->{AGENTS}->{list} =~ /^\s*RESTORE\s+(.*)$/)  {
			my $fn = MpmasQLFunctionLanguage->convert($1,$locvarhash);
			$total_agent_registry->clear();
			$agentregistry = SimpleObjectRegistry->recreate($fn);
			print "AGENTS restored from $fn\n";
		}	
		else {
			my @agentlist = MpmasQLFunctionLanguage->convert($disinfo->{AGENTS}->{list},$locvarhash);
			print "Creating ". scalar @agentlist. " new agents\n" if $verbose;
			$agentregistry = SimpleObjectRegistry->create();
			next CLULOOP if ( not defined $agentlist[0]);
			map { my $aid = $_; my $newagent = MasMapAgent->create(id => $aid,num_assigned => 0, num_members => 0, assigned => SimpleObjectRegistry->create(), memberlist => SimpleObjectRegistry->create() ); $agentregistry->add_objects($newagent); } @agentlist
		}
		for my $curAgent ($agentregistry->object_list) {
			my %locvars1  = %$locvarhash;
			$locvars1{agent_id} = $curAgent->id;
			$locvars1{agent} = $curAgent;
			for my $i (0..$#{$disinfo->{AGENTS}->{lines}}) {
				my ($field, $content) = split /\s*#=\s*/, $disinfo->{AGENTS}->{lines}->[$i];
				my ($keys, $fors) = MpmasQLFunctionLanguage->translate_and_interprete($field,\%locvars1);
				for my $j (0..$#{$keys}) {		
					my %locvars2  = %locvars1;
					$locvars2{attribute} = $fors->[$j];
					$curAgent->set_attribute($keys->[$j], MpmasQLFunctionLanguage->convert($content,\%locvars2) ||0 );
				}
			}
		}
		my $observationregistry = SimpleObjectRegistry->create();
		if (not defined $disinfo->{OBSERVATIONS}->{list} or $disinfo->{OBSERVATIONS}->{list} =~ /^\s*$/) {
			if ( is_in_list($disinfo->{TYPE}, [( qw(assign_by_cdf distribute_observation_list ) )]) ) {
					print STDERR "Error: You did not provide an OBSERVATIONS list for STEP $d\n";
					exit(1);
			}
		}
		elsif ($disinfo->{OBSERVATIONS}->{list} eq "NOTASSIGNED" ) {
			next unless ($notassigned->counter > 0);
			$observationregistry = dclone ($notassigned);
			$notassigned->clear ;
		}
		elsif ($disinfo->{OBSERVATIONS}->{list} =~ /^\s*RESTORE\s+(.*)$/)  {
			my $fn = MpmasQLFunctionLanguage->convert($1,$locvarhash);
			$notassigned->clear();
			$observationregistry = SimpleObjectRegistry->recreate($fn);
			print "OBSERVATIONS restored from $fn\n";
		}	
	
		else {
			my @observationlist = MpmasQLFunctionLanguage->convert($disinfo->{OBSERVATIONS}->{list},$locvarhash);
			next unless (defined $observationlist[0]);
			map { my $eid = $_; my $newobservation = SimpleObject->create(id => $eid); $observationregistry->add_objects($newobservation); } @observationlist
		}

		for my $curEndow ($observationregistry->object_list()) {
			my %locvars1  = %$locvarhash;
			#print $curEndow->get_attribute("id");
			$locvars1{observation_id} = $curEndow->get_attribute("id");
			$locvars1{observation} = $curEndow;
			for my $i (0..$#{$disinfo->{OBSERVATIONS}->{lines}}) {
				my ($field, $content) = split /\s*#=\s*/, $disinfo->{OBSERVATIONS}->{lines}->[$i];
				$|=1;
				my ($keys, $fors) = MpmasQLFunctionLanguage->translate_and_interprete($field,\%locvars1);
				for my $j (0..$#{$keys}) {		
					my %locvars2  = %locvars1;		
					$locvars2{attribute} = $fors->[$j];
					$curEndow->set_attribute($keys->[$j], MpmasQLFunctionLanguage->convert($content,\%locvars2) || 0 );
				}
			}
			#defaults:
			$curEndow->set_attribute("DEL_priority", INFTY) unless (defined $curEndow->get_attribute("DEL_priority"));
			$curEndow->set_attribute("CDF_priority", INFTY) unless (defined $curEndow->get_attribute("CDF_priority"));
			#$newagent->dump_me_to();
			if (defined $curEndow->get_attribute("OBS_multiply") and $curEndow->get_attribute("OBS_multiply") > 1) {
			      $observationregistry->delete_objects($curEndow);
			      foreach (1..$curEndow->get_attribute("OBS_multiply")) {
				    my $o = $curEndow->create_copy();
				    $o->set_attribute("id", $curEndow->get_attribute("id")* -100000 + $_);
				    $observationregistry->add_objects($o);
			      }
			}
		
		}
		
		if ($disinfo->{TYPE} eq "distribute_observation_list") {
			my @pre_potential_owners;
			## compile a first list of potential owners based on fixed and not observation_specific contraints
			AGENTLOOP: foreach my  $curAgent ($agentregistry->object_list) {
				my %loc2 = %$locvarhash;
				$loc2{agent_id} = $curAgent->get_attribute("id");
				$loc2{agent}	= $curAgent;
				push @pre_potential_owners, $curAgent if (evaluate_constraints($constraints, ["fixed"], [0,2], \%loc2, $verbose) == 1);
			}
			print $observationregistry->counter(). " endowment observations to distribute among ".scalar @pre_potential_owners. " agents\n" if $verbose;
			for my $curEndow (sort {$a->get_attribute("DEL_priority") <=> $b->get_attribute("DEL_priority")} $observationregistry->object_list()) {
				my $no_restriction_applied = 0;
				my %loc = %$locvarhash;
				$loc{observation_id} = $curEndow->get_attribute("id");
				$loc{observation}	= $curEndow;
				print "Trying to distribute ".$curEndow->get_attribute("DEL_count")." unit(s) of ". $curEndow->get_attribute("id")."\n" if ($verbose == 2);
				my @potential_owners;
				## compile list of potential owners based on fixed observation specific and all dynamic constraints if exist
				if (exists $constraints->{dynamic} or exists $constraints->{fixed}->{1} or exists $constraints->{fixed}->{3}) {
					AGENTLOOP: foreach my  $curAgent (@pre_potential_owners) {
						my %loc2 = %loc;
						$loc2{agent_id} = $curAgent->get_attribute("id");
						$loc2{agent}	= $curAgent;
						push @potential_owners, $curAgent if (evaluate_constraints($constraints, ["dynamic"], [0,1,2,3], \%loc2, $verbose) == 1 and evaluate_constraints($constraints, ["fixed"], [1,3], \%loc2, $verbose) == 1);
					}
				}
				else {
						$no_restriction_applied = 1;
						@potential_owners = @pre_potential_owners;
				}
				print scalar @potential_owners . " potential owners\n" if ($verbose == 2);
				## distribute as many as indicated by DEL_count attribute
				while ($curEndow->get_attribute("DEL_count") > 0) {
					if (@potential_owners <= 0) {
						print "\e[K \r\nNot assigned ".$curEndow->get_attribute("id"). " ".$curEndow->get_attribute("DEL_count")."\n" ;
						$notassigned->add_objects($curEndow);
						last;			
					}	
					###Random draw from potential agents
					my $draw = int(rand(@potential_owners));
					my $agent_to_assign_to = $potential_owners[$draw];
						
					my %loc2 = %loc;
					$loc2{agent_id} = $agent_to_assign_to->get_attribute("id");
					$loc2{agent}	= $agent_to_assign_to;
					do_assignment($disinfo, $agent_to_assign_to, $curEndow,  \%loc2, $verbose);
					do_hhmember_assignment($disinfo, $agent_to_assign_to, $curEndow,  \%loc2, $verbose);
					do_post_assignment($disinfo, $agent_to_assign_to, $curEndow,  \%loc2, $verbose);
					
					
					$curEndow->add_to_attribute("DEL_count", -1); 
					unless ( $curEndow->get_attribute("DEL_accumulate")) {
					    my @lower = @potential_owners[0..$draw -1 ];
					    my @upper = @potential_owners[$draw + 1 .. $#potential_owners];
					    @potential_owners = (@lower, @upper);
					}
					if (defined $curEndow->get_attribute("DEL_exclusive")  and $curEndow->get_attribute("DEL_exclusive") == 1) {
						my $index;
						if ($no_restriction_applied) {
						      $index = $draw;
						}
						else {
						 
						      foreach my $i (0..$#pre_potential_owners) {
							    if ($pre_potential_owners[$i]->id() == $loc2{agent_id}) {
								$index = $i; last;
							    }
						      }
						}
						my @plower = @pre_potential_owners[0..$index -1 ];
						my @pupper = @pre_potential_owners[$index + 1 .. $#pre_potential_owners];
						@pre_potential_owners = (@plower, @pupper);
					}
				}
			}
		}
		elsif ($disinfo->{TYPE} eq "assign_by_cdf") {
			my @prepre_potential_owners;
			## compile a first list of potential owners based on fixed and not observation_specific contraints
			AGENTLOOP: foreach my  $curAgent ($agentregistry->object_list) {
				my %loc2 = %$locvarhash;
				$loc2{agent_id} = $curAgent->get_attribute("id");
				$loc2{agent}	= $curAgent;
				push @prepre_potential_owners, $curAgent if (evaluate_constraints($constraints, ["fixed"], [0,2], \%loc2, $verbose) == 1);
			}
			my @standard;
			my @ensure;#
			print "No suitable agents left after applying fixed constraints\n" and next CLULOOP unless (scalar @prepre_potential_owners > 0);
			for my $curEndow ( $observationregistry->object_list()) {
				&prepare_icdf($curEndow);
				$curEndow->get_attribute("CDF_cdf")->qf_plot("@$clu"."_".$curEndow->get_attribute("id").".png") if ($verbose > 4 );
				$curEndow->get_attribute("CDF_cdf")->print_points() if ($verbose > 4);

				if (defined $curEndow->get_attribute("CDF_ensure_full") and $curEndow->get_attribute("CDF_ensure_full") ) {
				    push @ensure, $curEndow;
				}
				else {
				    push @standard, $curEndow;
				}
			}
			if (@ensure > 0) {
			    for my $curEndow (sort { $a->get_attribute("CDF_priority") <=> $b->get_attribute("CDF_priority") } @ensure) {
				my @pre_potential_owners;
			      ## compile a first list of potential owners based on fixed and not observation_specific contraints
			        AGENTLOOP: foreach my  $curAgent (@prepre_potential_owners) {
				      my %loc2 = %$locvarhash;
				      $loc2{agent_id} = $curAgent->get_attribute("id");
				      $loc2{agent}	= $curAgent;
				      push @pre_potential_owners, $curAgent if (evaluate_constraints($constraints, ["fixed"], [0,2], \%loc2, $verbose) == 1);
			        }
				my $lbatt =  $curEndow->get_attribute("CDF_part_lb_attribute");
				my $ubatt =  $curEndow->get_attribute("CDF_part_ub_attribute");
				my $binsize =  0.2;
				$binsize = $curEndow->get_attribute("CDF_ensure_binsize") if (defined $curEndow->get_attribute("CDF_ensure_binsize"));
				my %loc = %$locvarhash;
				$loc{observation_id} = $curEndow->get_attribute("id");
				$loc{observation}	= $curEndow;
				
				my $plevels = predraw_from_uniform(\@pre_potential_owners, $lbatt, $ubatt);
				my @levels = @$plevels;
				if ($curEndow->get_attribute("CDF_ensure_skip_zero") ) {
				    @levels =();
				    foreach (@$plevels) {
				      push @levels, $_ unless ($curEndow->get_attribute("CDF_cdf")->qf($_) == 0);
				    }
				}
				my $N = scalar @pre_potential_owners;
				#my $eps =  1.001 / $N;
				#print "EPS: $eps\n";
				#foreach my $i (1..$N) {
				#    push @levels, ($i  / $N) ;
				#}
				print $curEndow->get_attribute("id").": Distribute ". scalar @levels." levels among ". scalar @pre_potential_owners. " potential_owners\n" if $verbose;
				my $sign =   $curEndow->get_attribute("CDF_ensure_full_sorting") || 1;
				my $maxstage =  (defined  $curEndow->get_attribute("CDF_ensure_max_attempts")) ? ( $curEndow->get_attribute("CDF_ensure_max_attempts") - 1 ) : 0  ;
				my @missed_levels;
				my $last_missed;
				for my $stage (0 ..$maxstage) {
				    my @stack;
				    if ($stage == 0) {
					 if ($curEndow->get_attribute("CDF_ensure_full_shuffle")) {
					     @stack = @levels;
					 }
					 else {
					     @stack = sort {$b * $sign <=> $a * $sign} @levels;
					 }
				    }
				    else {
					@stack = @missed_levels;
					undef $last_missed;
					undef @missed_levels;
				    }
				    foreach my $l ( @stack ) {
					    $l = 1 if ($l > 1);
					    $curEndow->set_attribute("CDF_value_assigned", $curEndow->get_attribute("CDF_cdf")->qf($l) );
					    my $no_restriction_applied = 0;
					    my @potential_owners;
					    
					    ## compile list of potential owners based on fixed observation specific and all dynamic constraints if exist
					    if (defined $last_missed and $curEndow->get_attribute("CDF_value_assigned") == $last_missed) {
						    print "Skipping $l --> ".$curEndow->get_attribute("CDF_value_assigned")."\n";
					    }
					    elsif (defined $lbatt or defined $ubatt or exists $constraints->{dynamic} or exists $constraints->{fixed}->{1} or exists $constraints->{fixed}->{3} ) {
						    
						    AGENTLOOP: foreach my  $curAgent (@pre_potential_owners) {
							    if (defined $lbatt or defined $ubatt){
								my $lb = -1;
								my $ub = -1;
								$lb = $curAgent->get_attribute($lbatt)  / 100 if (defined $lbatt and defined $curAgent->get_attribute($lbatt) );
								$ub = $curAgent->get_attribute($ubatt)  / 100 if (defined $ubatt and defined $curAgent->get_attribute($ubatt) );
								if ($ub == -1 or $lb == -1) {
								    next unless ($stage == 1);
								    $ub = 1;
								    $lb = 0;  
								}
								next unless ($l >= $lb  and $l <= $ub );
							    }
							    my %loc2 = %loc;
							    $loc2{agent_id}	= $curAgent->get_attribute("id");
							    $loc2{agent}	= $curAgent;
							    $loc2{attempt}	= $stage +1;	
							    push @potential_owners, $curAgent if (evaluate_constraints($constraints, ["dynamic"], [0,1,2,3], \%loc2, $verbose) == 1 and evaluate_constraints($constraints, ["fixed"], [1,3], \%loc2, $verbose) == 1);
						    }
					    }
					    else {
							    $no_restriction_applied = 1;
							    @potential_owners = @pre_potential_owners;
					    }
					    #print $curEndow->get_attribute("id")." $l :".$curEndow->get_attribute("CDF_value_assigned"). " : ". scalar @potential_owners."\n";
					    
					    
					    if (@potential_owners <= 0) {
						    print "Not assigned ".$curEndow->get_attribute("id"). " $l -> ". $curEndow->get_attribute("CDF_value_assigned")."\n" if $verbose;
						    push @missed_levels, $l;
						    $last_missed = $curEndow->get_attribute("CDF_value_assigned");
						    next;			
					    }	
					    ###Random draw from potential agents
					    my $draw = int(rand(@potential_owners));
					    my $agent_to_assign_to = $potential_owners[$draw];
						    
					    my %loc2 = %loc;
					    $loc2{agent_id} = $agent_to_assign_to->get_attribute("id");
					    $loc2{agent}	= $agent_to_assign_to;
					    do_assignment($disinfo, $agent_to_assign_to, $curEndow,  \%loc2, $verbose);
					    do_hhmember_assignment($disinfo, $agent_to_assign_to, $curEndow,  \%loc2, $verbose);
					    do_post_assignment($disinfo, $agent_to_assign_to, $curEndow,  \%loc2, $verbose);
					    my $index;
					    if ($no_restriction_applied) {
						 $index = $draw;
					    }
					    else {
						foreach my $i (0..$#pre_potential_owners) {
						  if ($pre_potential_owners[$i]->id() == $loc2{agent_id}) {
						      $index = $i; last;
						  }
						}
					    }
					    my @plower = @pre_potential_owners[0..$index -1 ];
					    my @pupper = @pre_potential_owners[$index + 1 .. $#pre_potential_owners];
					    @pre_potential_owners = (@plower, @pupper);
				    }
				    print "Missed ".scalar @missed_levels. " levels.\n";
				}
			      }
			}
			if (@standard > 0) {
			      foreach my  $curAgent (@prepre_potential_owners) {
				      my %loc = %$locvarhash;
				      $loc{agent_id} = $curAgent->get_attribute("id");
				      $loc{agent}	= $curAgent;
				      
				      for my $curEndow (sort { $a->get_attribute("CDF_priority") <=> $b->get_attribute("CDF_priority") } $observationregistry->object_list()) {
					      
					      my %loc2 = %loc;
					      $loc2{observation_id} = $curEndow->get_attribute("id");
					      $loc2{observation}	= $curEndow;
					      my $lb = 0;
					      my $ub = 1;
					      my $lbatt =  $curEndow->get_attribute("CDF_part_lb_attribute");
					      my $ubatt =  $curEndow->get_attribute("CDF_part_ub_attribute");
					      $lb = $curAgent->get_attribute($lbatt)  / 100 if (defined $lbatt and defined $curAgent->get_attribute($lbatt) );
					      $ub = $curAgent->get_attribute($ubatt)  / 100 if (defined $ubatt and defined $curAgent->get_attribute($ubatt) );
		      
					      
					      #skip observationment if constraints for lowest assignment > 0 are not fulfilled
					      
					      $curEndow->set_attribute("CDF_value_assigned", $curEndow->get_attribute("CDF_cdf")->lowest_nonzero() );
					      next unless (evaluate_constraints($constraints, ["dynamic"], [0,1,2,3], \%loc2, $verbose) == 1 and evaluate_constraints($constraints, ["fixed"], [1,3], \%loc2, $verbose) == 1);
					      $curEndow->set_attribute("CDF_value_assigned", undef );
				      
					      my $maxit =  $curEndow->get_attribute("CDF_maxiterations") || 30;
					      my $it = 0;
					      while ($it++ < $maxit) {	
						      $curEndow->set_attribute("CDF_value_assigned", undef );
						      my $size_drawn = draw_value_from_cdf($curEndow->get_attribute("CDF_cdf"), $lb, $ub);
						      $curEndow->set_attribute("CDF_value_assigned", $size_drawn);
						      next unless (evaluate_constraints($constraints, ["dynamic"], [0,1,2,3], \%loc2, $verbose) == 1 and evaluate_constraints($constraints, ["fixed"], [1,3], \%loc2, $verbose) == 1);
						      do_assignment($disinfo, $curAgent, $curEndow,  \%loc2, $verbose);
						      do_hhmember_assignment($disinfo, $curAgent, $curEndow,  \%loc2, $verbose);
						      do_post_assignment($disinfo, $curAgent, $curEndow,  \%loc2, $verbose);
						      
						      last;
					      }	
					      $curEndow->set_attribute("CDF_value_assigned", undef );
					      
				      }
				  
			      }	
			}

		}	
		elsif ($disinfo->{TYPE} eq "fixed") {
			#fixed assignment for all agents that satisfy constraints
			foreach my  $curAgent ($agentregistry->object_list) {
				my %loc = %$locvarhash;
				$loc{agent_id} = $curAgent->get_attribute("id");
				$loc{agent}	= $curAgent;
				next unless (evaluate_constraints($constraints, [("dynamic", "fixed")], [(0,1,2,3)], \%loc, $verbose) == 1 );
				do_assignment($disinfo, $curAgent, undef,  \%loc, $verbose);
				do_hhmember_assignment($disinfo, $curAgent, undef,  \%loc, $verbose);
				do_post_assignment($disinfo, $curAgent, undef,  \%loc, $verbose);
			}
		}
		elsif ($disinfo->{TYPE} eq "fixed_repeated") {
			#fixed assignment for all agents that satisfy constraints
			foreach my  $curAgent ($agentregistry->object_list) {
				my %loc = %$locvarhash;
				$loc{agent_id} = $curAgent->get_attribute("id");
				$loc{agent}	= $curAgent;
				next unless (evaluate_constraints($constraints, [("fixed")], [(0,1,2,3)], \%loc, $verbose) == 1 );
				while (evaluate_constraints($constraints, [("dynamic")], [(0,1,2,3)], \%loc, $verbose) == 1) { 
				    do_assignment($disinfo, $curAgent, undef,  \%loc, $verbose);
				    do_hhmember_assignment($disinfo, $curAgent, undef,  \%loc, $verbose);
				    do_post_assignment($disinfo, $curAgent, undef,  \%loc, $verbose);
				}
			}
		}
		elsif ($disinfo->{TYPE} eq "fit_into_subarea") {
		      
		      if (not defined $disinfo->{SUBAREAS}->{list} or $disinfo->{SUBAREAS}->{list} =~ /^\s*$/) {
			      unless ($subarearegistry->counter > 0){
				  print STDERR "Error: You did not provide a SUBAREA list for STEP $d\n";
				  exit(1); 
			      }
		      }
		      
		      elsif ($disinfo->{SUBAREAS}->{list} =~ /^\s*RESTORE\s+(.*)$/)  {
			      my $fn = MpmasQLFunctionLanguage->convert($1,$locvarhash);
			      $subarearegistry = SimpleObjectRegistry->recreate($fn);
			      print "SUBAREAS restored from $fn\n";
		      }
		      else {
			      $subarearegistry->clear();
			      my @subarealist = MpmasQLFunctionLanguage->convert($disinfo->{SUBAREAS}->{list},$locvarhash);
			      #next unless (defined $subarealist[0]);
			      map { my $eid = $_; my $subarea = SimpleObject->create(id => $eid); $subarearegistry->add_objects($subarea); } @subarealist
		      }
		      my %obs_in_area;
		      for my $curArea ($subarearegistry->object_list()) {
			      $obs_in_area{$curArea->get_attribute("id")} = SimpleObjectRegistry->create();
			      my %locvars1  = %$locvarhash;
			      $locvars1{area_id} = $curArea->get_attribute("id");
			      $locvars1{area} = $curArea;
			      for my $i (0..$#{$disinfo->{SUBAREAS}->{lines}}) {
				      my ($field, $content) = split /\s*#=\s*/, $disinfo->{SUBAREAS}->{lines}->[$i];
				      $|=1;
				      my ($keys, $fors) = MpmasQLFunctionLanguage->translate_and_interprete($field,\%locvars1);
				      for my $j (0..$#{$keys}) {		
					      my %locvars2  = %locvars1;		
					      $locvars2{attribute} = $fors->[$j];
					      $curArea->set_attribute($keys->[$j], MpmasQLFunctionLanguage->convert($content,\%locvars2) || 0 );
				      }
			      }
		      }		
		      
		      for my $o ($observationregistry->object_list() ) {
			      my $x = $o->get_attribute("FIT_subarea");
			      
			      print STDERR "Observation ". $o->get_attribute("id"). " is missing the FIT_subarea attribute!\n" and exit(1) unless (defined $x);
			      print STDERR "Observation ". $o->get_attribute("id"). " is supposed to lie in subarea $x, which does not exist!\n" and exit(1) unless (exists $obs_in_area{$x});
			      $obs_in_area{$x}->add_objects($o);
			
		      }
		      print STDERR "SUBAREA_CONSTRAINED_FIELDS not defined or empty\n" and exit(1) unless defined ($disinfo->{SUBAREA_CONSTRAINED_FIELDS});
		      my @areafields = MpmasQLFunctionLanguage->convert($disinfo->{SUBAREA_CONSTRAINED_FIELDS},{});
		      print "@areafields\n" if $verbose > 1;

		      my @agents_without_match;
		      my @agentlist = sort {$a->get_attribute("FIT_priority") <=> $b->get_attribute("FIT_priority")} $agentregistry->object_list();
		      MTCAGLOOP: while  (@agentlist) {
			    my $curAgent = shift @agentlist;
			    $curAgent->set_attribute("FIT_split",0);
			    my %loc = %$locvarhash; 
			    $loc{agent_id} = $curAgent->get_attribute("id");
			    $loc{agent}	= $curAgent;
			    if (defined $disinfo->{DO_NOT_MATCH_IF}) {
				for my $j (0..$#{$disinfo->{DO_NOT_MATCH_IF}->{lines}}) {
					      my $L = $disinfo->{DO_NOT_MATCH_IF}->{lines}->[$j];
					      next if ($L =~ /^\s*$/);
					      next MTCAGLOOP if MpmasQLFunctionLanguage->convert($L, \%loc ) ;	
				}
			    }
			    my @potential_observations;
			    my $predef_s = $curAgent->get_attribute("FIT_subarea");
			    unless ( defined $predef_s ) {
				  $curAgent->set_attribute("FIT_second", 0);
				  SUBSEARCH: for my $s ($subarearegistry->object_list()) {
				      for my $f (@areafields) {
					  print STDERR "Attribute $f not found for agent ". $curAgent->id."\n" and exit(1) unless (defined  $curAgent->get_attribute("$f"));
					  print STDERR "Attribute $f not found for subarea ". $s->get_attribute("id")."\n"  and exit(1) unless (defined  $s->get_attribute("$f"));
					  next SUBSEARCH if ($curAgent->get_attribute("$f") > $s->get_attribute("$f") );
				      }  
				      push @potential_observations, $obs_in_area{ $s->get_attribute("id") }->object_list;
				  }
			    }
			    else {
				   $curAgent->add_to_attribute("FIT_second", 1);
				   push @potential_observations, $obs_in_area{ $predef_s }->object_list;

			    }
			    if (scalar @potential_observations == 0 ) {
				push (@agents_without_match, $curAgent) and print "No match for agent ". $curAgent->id." (no subarea with capacity)\n" unless $curAgent->get_attribute("FIT_second");
				next ; 
			    }
			    my %agglikelihoods;
			    my $sum_of_likelihoods = 0;
			    
			    
			    foreach my $curEndow (@potential_observations) {
						    my %loc2 = %loc;
						    $loc2{observation_id} = $curEndow->get_attribute("id");
						    $loc2{observation}	= $curEndow;
						    #calculate likelihoods
						    $agglikelihoods{ $curEndow->get_attribute("id")} = 1;
						    
						    for my $j (0..$#{$disinfo->{LIKELIHOODS}->{lines}}) {
							  my $L = $disinfo->{LIKELIHOODS}->{lines}->[$j];
							  next if ($L =~ /^\s*$/);
							  my $res = MpmasQLFunctionLanguage->convert($L, \%loc2 ) ;	
							  
							  $agglikelihoods{ $curEndow->get_attribute("id")} *= $res;
						    }
						    $sum_of_likelihoods += $agglikelihoods{$curEndow->get_attribute("id")};
				}
			my @obs_drawn;
			if ($sum_of_likelihoods > 0) {
				  my $curEndow = fit_into_subarea_draw (\@potential_observations, \%agglikelihoods, $observationregistry, $sum_of_likelihoods);
				  push @obs_drawn, $curEndow;
				  $obs_in_area{$curEndow->get_attribute("FIT_subarea")}->delete_objects($curEndow);
			}
		        elsif (defined $disinfo->{AGENT_SPLIT_ATTRIBUTES}) {
				my @splitattributes = MpmasQLFunctionLanguage->convert($disinfo->{AGENT_SPLIT_ATTRIBUTES}, {} );
				my $matchfound = 0;
				my @agent_parts = ($curAgent);
				
				SPLITLOOP: while (scalar @agent_parts <= $disinfo->{MAX_SPLIT}) {
				      print "Split round ". scalar @agent_parts ."\n" if ($verbose > 2);
				      my $smallsh = rand(.5);
				      my $largesh = 1 -$smallsh;
				      @agent_parts = sort {$b->get_attribute($splitattributes[0]) <=> $a->get_attribute($splitattributes[0]) } @agent_parts;
				      my $to_split = shift @agent_parts;
				      my $largepart = $to_split->create_copy();
				      my $smallpart = $to_split->create_copy();
				      $largepart->set_attribute("FIT_split",1);
				      $smallpart->set_attribute("FIT_split",1);
				      foreach (@splitattributes) {
					  $largepart->set_attribute($_, $to_split->get_attribute($_) * $largesh);
					  $smallpart->set_attribute($_, $to_split->get_attribute($_) * $smallsh);
				      }
				      push @agent_parts, $largepart;
				      push @agent_parts, $smallpart;
				      #print "Parts:";
				      #foreach (@agent_parts) {print $_->get_attribute("water")." ";};
				      #print "\n";
				      my %agglikelihoods;
				      my %sum_of_likelihoods_per_sector_per_part;
			    
			    
				      foreach my $curEndow (@potential_observations) {
						
						my %loc2 = %loc;
						$loc2{observation_id} = $curEndow->get_attribute("id");
						$loc2{observation}	= $curEndow;
						#calculate likelihoods
						
						foreach my $p (0..$#agent_parts) {
							$agglikelihoods{$p}->{ $curEndow->get_attribute("id")} = 1;
							my %loc3 = %loc2;
							$loc3{agent_id} = $agent_parts[$p]->get_attribute("id");
							$loc3{agent}	= $agent_parts[$p];
							for my $j (0..$#{$disinfo->{LIKELIHOODS}->{lines}}) {
							      my $L = $disinfo->{LIKELIHOODS}->{lines}->[$j];
							      next if ($L =~ /^\s*$/);
							      my $res = MpmasQLFunctionLanguage->convert($L, \%loc3 ) ;	
							      
							      $agglikelihoods{$p}->{ $curEndow->get_attribute("id")} *= $res;
							}
							my $sec =  $curEndow->get_attribute("FIT_subarea");
							$sum_of_likelihoods_per_sector_per_part{$sec}->{$p} = 0 if (not exists $sum_of_likelihoods_per_sector_per_part{$sec}->{$p});
							$sum_of_likelihoods_per_sector_per_part{$sec}->{$p} += $agglikelihoods{$p}->{$curEndow->get_attribute("id")};
						}
				      }
				      my @secs_with_likelihood_for_all;
				      #print Dumper(\%sum_of_likelihoods_per_sector_per_part);
				      O1: foreach my $sec (keys %sum_of_likelihoods_per_sector_per_part) {
					    O2: foreach my $p (keys %{$sum_of_likelihoods_per_sector_per_part{$sec}} ) {  
							next O1 if  ($sum_of_likelihoods_per_sector_per_part{$sec}->{$p} <= 0);
					    }
					     push @secs_with_likelihood_for_all, $sec;
				      }
				      next SPLITLOOP unless (scalar @secs_with_likelihood_for_all > 0);
				      print "Sectors: @secs_with_likelihood_for_all\n" if ($verbose > 2);
				      my @obs_in_sec_with_likelihood_for_all;
				      my %sum_of_likelihoods_per_part;
				      for my $sec (@secs_with_likelihood_for_all){
					      push @obs_in_sec_with_likelihood_for_all, $obs_in_area{ $sec }->object_list;
					      
				      }
				      my $firstEndow = fit_into_subarea_draw (\@obs_in_sec_with_likelihood_for_all ,$agglikelihoods{0}, $observationregistry, undef());
				      my $sec =  $firstEndow->get_attribute("FIT_subarea");
				      push @obs_drawn, $firstEndow;
				      $obs_in_area{$sec}->delete_objects($firstEndow);
				      
				      foreach my $p (1..$#agent_parts) {
					    my $othEndow = fit_into_subarea_draw ( [$obs_in_area{ $sec }->object_list] ,$agglikelihoods{$p}, $observationregistry, undef());
					    if (not defined $othEndow) {  
					        $obs_in_area{$sec}->add_objects(@obs_drawn);
						@obs_drawn = ();
						next SPLITLOOP
					    }
					    
					    push @obs_drawn, $othEndow;
					    $obs_in_area{$sec}->delete_objects($othEndow);
				      }
				      $matchfound = 1;
				      last SPLITLOOP;
				}
				unless ($matchfound) {
					push (@agents_without_match, $curAgent) and print "No match for agent ". $curAgent->id." (reached MAX_SPLIT)\n" if (not $curAgent->get_attribute("FIT_second")); 
					  next MTCAGLOOP;
				}
			}
			else {
				push (@agents_without_match, $curAgent) and print "No match for agent ". $curAgent->id." (no positive likelihoods in subareas with capacity)\n" if (not $curAgent->get_attribute("FIT_second")); 
				next MTCAGLOOP;
			}
			foreach my $o (0..$#obs_drawn) {
				  my $curEndow = $obs_drawn[$o];
				  my %loc2 = %loc;
				  $loc2{observation_id} = $curEndow->get_attribute("id");
				  $loc2{observation}	= $curEndow;
				  do_assignment($disinfo, $curAgent, $curEndow,  \%loc2, $verbose);
				  do_hhmember_assignment($disinfo, $curAgent, $curEndow,  \%loc2, $verbose);
				  do_post_assignment($disinfo, $curAgent, $curEndow,  \%loc2, $verbose);
				  $curAgent->set_attribute( "FIT_match_$o", $loc2{observation_id}  );
				  my $s = $curEndow->get_attribute("FIT_subarea");
				  if ($o == 0) {
					
					$curAgent->set_attribute("FIT_subarea", $s);
					my $subarea = $subarearegistry->get_object_by_id($s);
					for my $f (@areafields) {
						$subarea->add_to_attribute("$f", -1 *  $curAgent->get_attribute("$f"));
					}
					
				  }
				  print join("\t",$curAgent->id, $curEndow->get_attribute("id"), $s)."\n" if $verbose;
				  
			   }
		      } ##END MTCAGLOOP
		      print "Remaining capacities in subareas:\n";
		      foreach my $s (sort {$a->get_attribute("id") <=>  $b->get_attribute("id")} $subarearegistry->object_list  ) {
			    print $s->get_attribute("id");
			    for my $f (@areafields) {
				     print "\t".$s->get_attribute("$f");
			    }
			    print "\n";
			    $notassigned->add_objects($obs_in_area{$s->get_attribute("id")}->object_list() )
			  
		      }
		      


		}
		elsif ($disinfo->{TYPE} eq "match" or $disinfo->{TYPE} eq "match_cdf") {
			
			my %likelihood_lines;
			for my $j (0..$#{$disinfo->{LIKELIHOODS}->{lines}}) {
				my $i = $disinfo->{LIKELIHOODS}->{lines}->[$j];
				my ($Ldim, $L) = split (/\s*#=\s*/, $i, 2 );
				my $translated = MpmasQLFunctionLanguage->translate_to_perl($L) ;
				$likelihood_lines{$Ldim} = $translated;
			}
			my @all_agents;
			AGENTLOOP: foreach my  $curAgent ($agentregistry->object_list) {
				my %loc2 = %$locvarhash;
				$loc2{agent_id} = $curAgent->get_attribute("id");
				$loc2{agent}	= $curAgent;
				push @all_agents, $curAgent if (evaluate_constraints($constraints, ["fixed"], [0,2], \%loc2, $verbose) == 1);
			}
			
			if (defined $disinfo->{SORT} ) {
			    @all_agents = sort {$b->get_attribute($disinfo->{SORT}) <=> $a->get_attribute($disinfo->{SORT}) } @all_agents;
			}
			my %match_regions;
			my $parent_observation;
			if ($disinfo->{TYPE} eq "match_cdf") {
				my @xx = $observationregistry->object_list();
				my $curEndow =  $xx[0];
				$parent_observation =  $xx[0];
				&prepare_icdf($curEndow);
				my $lbatt =  $curEndow->get_attribute("CDF_part_lb_attribute");
				my $ubatt =  $curEndow->get_attribute("CDF_part_ub_attribute");
				my $binsize =  0.2;
				$binsize = $curEndow->get_attribute("CDF_ensure_binsize") if (defined $curEndow->get_attribute("CDF_ensure_binsize"));
				my %bins;
				for my $mr (1..1/$binsize) {
				    next unless ( $curEndow->get_attribute("CDF_cdf")->qf( ($mr + 1) * $binsize ) > 0 or $mr == 1 / $binsize);
				    $bins{$binsize * 100 *$mr} = $binsize * 100 *$mr;
				    $match_regions{$binsize * 100 *$mr}->{agents} = [];
				    $match_regions{$binsize * 100 *$mr}->{observations} = [];
				}
				for my $ag (@all_agents) {
					push @{$match_regions{ determine_class_by_upperbound(\%bins,  $ag->get_attribute($ubatt) ) }->{agents} }, $ag;
				}
				my $levels = predraw_from_uniform(\@all_agents, $lbatt, $ubatt);	
				my $c = 0;
				foreach my $l (@$levels) {
					    $l = 1 if ($l > 1);
					    my $newobservation = SimpleObject->create(id => $c++); 
					    $newobservation->set_attribute("CDF_quantile",$l);
					    $newobservation->set_attribute("CDF_value_assigned", $curEndow->get_attribute("CDF_cdf")->qf($l) );
					    push @{ $match_regions{determine_class_by_upperbound(\%bins, $l*100)}->{observations} }, $newobservation;
				}	    
			      
			      
			}
			else {
			      $match_regions{0}->{observations} = $observationregistry->object_list();
			      $match_regions{0}->{agents} = @all_agents;
			}
			
			foreach my $mr (sort {$b <=> $a} keys %match_regions) {
				my @observations = @{ $match_regions{$mr}->{observations} };
				my @sorted_agents = @{ $match_regions{$mr}->{agents} };
				my %likelihoods;
				my %aggloglikelihoods;
				print "Match region $mr\n";
				print "Obs: ". scalar @observations. ", Ags: ". scalar @sorted_agents ."\n";
				print "Calculating likelihoods...\n";
				foreach my  $curAgent (@sorted_agents) {
					my %loc = %$locvarhash; 
					$loc{agent_id} = $curAgent->get_attribute("id");
					$loc{agent}	= $curAgent;
					OLOOP: foreach my $curEndow (@observations) {
						my %loc2 = %loc;
						$loc2{observation_id} = $curEndow->get_attribute("id");
						$loc2{observation}	= $curEndow;
						#calculate likelihoods
						$aggloglikelihoods{ $curEndow->get_attribute("id")}-> {$curAgent->get_attribute("id")} = 0;
						for my $j (keys %likelihood_lines) {
						      my $res = MpmasQLFunctionLanguage->convert($likelihood_lines{$j}, \%loc2 , 1) ;	
						      $likelihoods{$j}->{ $curEndow->get_attribute("id")}-> {$curAgent->get_attribute("id")} = $res;
							if ( $res == 0) {
							    $aggloglikelihoods{ $curEndow->get_attribute("id")}-> {$curAgent->get_attribute("id")} = undef;
							    last;
							}
							$aggloglikelihoods{ $curEndow->get_attribute("id")}-> {$curAgent->get_attribute("id")} += log ($res);
						}
					}
				}
				if ($verbose >= 3) {
				      foreach my $l (keys %likelihoods) {
					    foreach my $i (keys %{$likelihoods{$l}} ) { 
						    foreach my $j (keys %{$likelihoods{$l}->{$i}} ) { 
							  print join ("\t", $l, $i, $j, $likelihoods{$l}->{$i}->{$j}) ."\n"; 
						    }
					    }
				      }
				}
				my %sumlikelihood;
				foreach my $i (keys %aggloglikelihoods ) { 
					foreach my $j (keys %{$aggloglikelihoods{$i}} ) { 
					    $sumlikelihood{$j} +=  exp($aggloglikelihoods{$i}->{$j}) if ( defined $aggloglikelihoods{$i}->{$j});
					}
				}
				if ($verbose >= 2) {
					  print "obs\tagent\tlogL\tL\tP\n";
					  foreach my $i (keys %aggloglikelihoods ) { 
						  foreach my $j (keys %{$aggloglikelihoods{$i}} ) { 
						      if ( defined $aggloglikelihoods{$i}->{$j}) {
							print join ("\t", $i, $j, $aggloglikelihoods{$i}->{$j}, exp($aggloglikelihoods{$i}->{$j}), exp($aggloglikelihoods{$i}->{$j}) /$sumlikelihood{$j} ) ."\n"; 
						      }
						      else {
							print join ("\t", $i, $j, "N/A", 0, 0 ) ."\n"; 
						      }
						  }
					  }
				}
				if ($disinfo->{MATCHING} eq "max_milp") {
				      my $milp = MpmasMILP->adhoc_create();
				      if (defined $disinfo->{PERTURB_EQPROB} and $disinfo->{PERTURB_EQPROB} eq "1") {
				      

				      }
				      my $failflag = 0;
				      my @agents_left;
				      foreach my  $curAgent (@sorted_agents) {
					    my @potentials = grep{ defined $aggloglikelihoods{$_}->{$curAgent->id} }  keys %aggloglikelihoods;
					    print $curAgent->id."\t@potentials\n" if $verbose;
					    if ( scalar @potentials == 0) {
						    $failflag = 1 if (uc($disinfo->{ON_AGENTS_LEFT}) eq "FAIL") ;
						    push @agents_left , $curAgent->id;
						    next if (uc($disinfo->{ON_AGENTS_LEFT}) eq "FAIL" or uc($disinfo->{ON_AGENTS_LEFT}) eq "REPORT");
					    }
					    
					    $milp->adhoc_add_constraint(
						    "const_agent_".$curAgent->id,{
						    name => "const_agent_".$curAgent->id,
						    eqtype => 3,
						    value => 1});
				      }
				    if ($failflag) {
					  print STDERR "There are agents without any potential match among the observations:\n";
					  print join(",", @agents_left)."\n";
					  exit(1);
				    } 
				    elsif ( scalar @agents_left > 0 ) {
					  print "There are agents without any potential match among the observations:\n";
					  print join(",", @agents_left)."\n";
				    }
				      foreach my $i (keys %aggloglikelihoods ) { 
					    $milp->adhoc_add_constraint(
						    "const_obs_$i",{
						    name => "const_obs_$i",
						    eqtype =>  1,
						    value => 1});
					    foreach my $j (keys %{$aggloglikelihoods{$i}} ) { 
						if ( defined $aggloglikelihoods{$i}->{$j}) {
						      $milp->adhoc_add_activity(
							      $i."_".$j,
							      {name => $i."_".$j,
							      integ => 1,
							      ubound => 1,
							      orow =>  $aggloglikelihoods{$i}->{$j},
							      "const_obs_$i" => 1,
							      "const_agent_$j" => 1}
						      )
						}
						elsif (defined $disinfo->{"ALLOW_BADMATCHES"} and $disinfo->{"ALLOW_BADMATCHES"} == 1) {
						      $milp->adhoc_add_activity(
							      $i."_".$j,
							      {name => $i."_".$j,
							      integ => 1,
							      ubound => 1,
							      orow =>  -1 * 10**15,
							      "const_obs_$i" => 1,
							      "const_agent_$j" => 1}
						      )
						}
					    }
				      }
				      print "Started solving of maximization problem!\n";
				      $milp->adhoc_solve($milpcheck,1) or print STDERR "Solution failed!\n".$milp->adhoc_status_message()."\n"   and exit(1);
				      
				      print "agent\thhd\tchoice\n" if ($verbose);
				      foreach my $i ($milp->activities) {
					    if ($i->get_attribute("solution") > 0 ) {
						$i->id =~ /(\d+)_(\d+)/;  
						print join("\t", $2,$1, $i->get_attribute("solution"))."\n" if ($verbose);
						my $agent_to_assign_to = $agentregistry->get_object_by_id($2);
						my $curEndow = $observationregistry->get_object_by_id($1);
						my %loc2 = %$locvarhash; 
						$loc2{observation_id} = $curEndow->get_attribute("id");
						$loc2{observation}	= $curEndow;
						$loc2{agent_id} = $agent_to_assign_to->get_attribute("id");
						$loc2{agent}	= $agent_to_assign_to;
						do_assignment($disinfo, $agent_to_assign_to, $curEndow,  \%loc2, $verbose);
						do_hhmember_assignment($disinfo, $agent_to_assign_to, $curEndow,  \%loc2, $verbose);
						do_post_assignment($disinfo, $agent_to_assign_to, $curEndow,  \%loc2, $verbose);
					    }
				      }

				}
				elsif ($disinfo->{MATCHING} eq "max_hungarian") {
				      
				      my @match_matrix;
				      $disinfo->{USE_ZEROL_PENALTY} = INFTY unless (defined $disinfo->{USE_ZEROL_PENALTY});
				      my @match_column = sort {$a <=> $b } keys %aggloglikelihoods;
				      my $cct = 0;
				      foreach my $i (@match_column ) { 
					    my $rct = 0;
					    
					    foreach my $j (@sorted_agents ) { 
						  $match_matrix[$rct++][$cct] = defined $aggloglikelihoods{$i}->{$j->id} ? -1 * $aggloglikelihoods{$i}->{$j->id} : $disinfo->{USE_ZEROL_PENALTY} ; 
					    }
					    $cct++;
				      }
				      if (defined $disinfo->{OBS_NONUSE_PENALTY_ATT}) {
					  my $diff =  scalar (keys (%aggloglikelihoods)) - scalar (@sorted_agents) ;
					  if ($diff > 0) {
						for my $j (0..$diff-1) {
						      for my $i (0..$#match_column) {
							my $obs = $observationregistry->get_object_by_id($match_column[$i]);
							my $val = $obs->get_attribute("$disinfo->{OBS_NONUSE_PENALTY_ATT}");
							die "Observation ". $obs->get_attribute("id"). " does not have a ". $disinfo->{OBS_NONUSE_PENALTY_ATT}. " attribute!\n" unless (defined $val);
							$match_matrix[ scalar (@sorted_agents) + $j][$i] = $val;
						      }
						}
					  }
				      }
				      my @match_result;
				      print "Start matching using Hungarian Algorithm (Kuhn, Munkres)\n";
				      use Algorithm::Munkres ();
				      Algorithm::Munkres::assign( \@match_matrix, \@match_result);
				      foreach my $j (0..$#sorted_agents) {
					      print $sorted_agents[$j]->id."\t$match_column[$match_result[$j]]\t".$aggloglikelihoods{$match_column[$match_result[$j]]}->{$sorted_agents[$j]->id}."\n" if $verbose;
					      my $agent_to_assign_to = $agentregistry->get_object_by_id($sorted_agents[$j]->id);
						my $curEndow = $parent_observation;
						my $lev;
						if ($disinfo->{TYPE} eq "match_cdf") {
						    for my $o (@observations) {
							if ( $match_column[$match_result[$j]] == $o->id) {
							      $lev  = $o ;
							      last;
							}
							
						    }
						    print "Error: Matching observation $match_column[$match_result[$j]] not found!\n" and exit(1) unless (defined $lev);
						    $curEndow->set_attribute("CDF_value_assigned", $lev->get_attribute("CDF_value_assigned") );
						}
						else {
						    $curEndow  = $observationregistry->get_object_by_id($match_column[$match_result[$j]]);
						}
						my %loc2 = %$locvarhash; 
						$loc2{observation_id} = $curEndow->get_attribute("id");
						$loc2{observation}	= $curEndow;
						$loc2{agent_id} = $agent_to_assign_to->get_attribute("id");
						$loc2{agent}	= $agent_to_assign_to;
						do_assignment($disinfo, $agent_to_assign_to, $curEndow,  \%loc2, $verbose);
						do_hhmember_assignment($disinfo, $agent_to_assign_to, $curEndow,  \%loc2, $verbose);
						do_post_assignment($disinfo, $agent_to_assign_to, $curEndow,  \%loc2, $verbose);
				      }
				      print "Unused observations:\n" if $verbose;
				      foreach my $j ($#sorted_agents+1 .. $#match_result) {
					      print $match_column[$match_result[$j]]."\n" if $verbose;
				      }
				}
			}
		}
		elsif ($disinfo->{TYPE} eq "maps_artificial") {
		        my %loc = %$locvarhash;
  
			my $cellsize	=	MpmasQLFunctionLanguage->convert($disinfo->{CELLSIZE},{});
	
			my @attr_to_distribute;
			foreach ( @{$disinfo->{SOILCODES}->{lines}}) {
				my ($attr,$suit) =  split /\s*#=\s*/, $_, 2;
				my ($attlist, $attfor) =  MpmasQLFunctionLanguage->translate_and_interprete($attr,\%loc);
				
				foreach my $k (0..$#{$attlist}) { 
						my %hash;
						$hash{field} = $attlist->[$k];
						my %loc2 = %loc;
						$loc2{attribute} = $attfor->[$k];
						$hash{soil_code} =  MpmasQLFunctionLanguage->convert($suit,\%loc2);

						push @attr_to_distribute, \%hash;
				} ;
			}

			my @soilarray;
			my @farmsteadarray;
			my @propertyarray;
			my $cellpointer = 0;
			for my $curAgent ($agentregistry->object_list ) {
				my $nplots;
				my @slist;
				for my $s (@attr_to_distribute) {
				    my $n_of_s = sprintf("%.0f",$curAgent->get_attribute($s->{field}));
				    push @soilarray, ($s->{soil_code}) x $n_of_s; 
				    $nplots += $n_of_s;
				}
			    if ($nplots > 0) { push @farmsteadarray, $curAgent->id, (-1) x ($nplots -1) };
			    push @propertyarray, ($curAgent->id) x $nplots;
			}
	
			my $soilmap = EnhancedGrid->create_map_from_sequence(\@soilarray, $disinfo->{MAPHEIGTH},-1, "SOILMAP", $cellsize);
			my $fstmap = EnhancedGrid->create_map_from_sequence(\@farmsteadarray, $disinfo->{MAPHEIGTH},-1, "FARMSTEADMAP", $cellsize);
			my $propertymap = EnhancedGrid->create_map_from_sequence(\@propertyarray, $disinfo->{MAPHEIGTH},-1, "PROPERTYMAP", $cellsize);
			$mapregistry->add_or_replace_objects($soilmap);
			$mapregistry->add_or_replace_objects($fstmap);
			$mapregistry->add_or_replace_objects($propertymap);
		}
		elsif ($disinfo->{TYPE} eq "make_secondary_maps_for_mpmas") {
			my $farmsteadmap= $mapregistry->get_object_by_id("FARMSTEADMAP");
			my $propertymap= $mapregistry->get_object_by_id("PROPERTYMAP");

			my %reclass;
			for my $curAgent ($agentregistry->object_list ) {
				my $nplots;
				my @slist;
				for my $s ( qw(MASCluster MASPopulation MASNetwork MASSector MASCatchment) ) {
				    my $c = $curAgent->get_attribute($s);
				    print STDERR "Error when making Mpmas secondary maps: Attribute $s undefined for agent $curAgent\n" and exit(1) if (not defined $c);
				    print STDERR "Error when making Mpmas secondary maps: Attribute $s is not numeric (= $c) for agent $curAgent\n" and exit(1) if (not looks_like_number($c));
				    $reclass{$s}->{$curAgent->id} = $c;
				}
			}
			my $clumap = $farmsteadmap->create_copy_and_reclassify($reclass{"MASCluster"}, -1);
			my $popmap = $farmsteadmap->create_copy_and_reclassify($reclass{"MASPopulation"}, -1);
			my $netmap = $farmsteadmap->create_copy_and_reclassify($reclass{"MASNetwork"}, -1);
			my $secmap = $propertymap->create_copy_and_reclassify($reclass{"MASSector"}, -1);
			my $catchmap = $propertymap->create_copy_and_reclassify($reclass{"MASCatchment"}, -1);
			$clumap->set_id("CLUSTERMAP");
			$popmap->set_id("POPULATIONMAP");
			$netmap->set_id("NETWORKMAP");
			$secmap->set_id("SECTORMAP");
			$catchmap->set_id("CATCHMENTMAP");
			$mapregistry->add_or_replace_objects($clumap, $popmap, $netmap, $secmap, $catchmap);
		}
		elsif ($disinfo->{TYPE} eq "maps_assign_farmstead") {
			my %loc = %$locvarhash;
			print STDERR "Error: You need to specify MAP for maps_assign_farmstead\n" and exit(1) unless (exists $disinfo->{MAP});
			my $smapname = MpmasQLFunctionLanguage->convert($disinfo->{MAP},\%loc);
			my $suitabilitymap = $mapregistry->get_object_by_id($smapname);
			print STDERR "Unknown MAP: $smapname\n" and exit(1) unless (defined $suitabilitymap);
			my $farmsteadmap= $mapregistry->get_object_by_id("FARMSTEADMAP");
			unless (defined $farmsteadmap ) {
				$farmsteadmap = $suitabilitymap->create_alike(-1);
				$farmsteadmap->reset_nodata_to(-1);
				$farmsteadmap->set_id("FARMSTEADMAP");
			}
			
			
			my %agentgroups;
			my %suits;
			if ($disinfo->{SUITABLE} =~ /#this\s*\(\s*agent/ ) { 
				
				for my $curAgent ($agentregistry->object_list ) {
					my %loc2 = %loc;
					$loc2{agent_id} = $curAgent->get_attribute("id");
					$loc2{agent}	= $curAgent;
					my @suitable = sort {$a <=> $b} MpmasQLFunctionLanguage->convert($disinfo->{SUITABLE},\%loc2);
					my $suitcode = join ("_", @suitable);
					push @{$agentgroups{$suitcode}}, $curAgent ;
					$suits{$suitcode} = \@suitable;
					#print "Agent $loc{agent_id} , suitcode $suitcode\n" if $verbose;
				}
			}
			
			else {
				my @suitable = sort {$a <=> $b} MpmasQLFunctionLanguage->convert($disinfo->{SUITABLE},\%loc);
				my $suitcode = join ("_", @suitable);
				$suits{$suitcode} = \@suitable;
				$agentgroups{$suitcode}=[( $agentregistry->object_list() )];
			}
			for my $suitcode (keys %agentgroups) {
				print "Distributing farmsteads to ". scalar ( @{ $agentgroups{$suitcode} }) ." agents of suitability group $suitcode\n" if $verbose;
				my $region = MasMapRegion->create();
				$region->add_agents(@{ $agentgroups{$suitcode}});
				$region->set_farmsteadmap($farmsteadmap);
				$region->distribute_farmsteads($suitabilitymap, $suits{$suitcode}, $verbose) or print STDERR "Could not assign a farmstead to all agents in $suitcode\n" and exit(1);
				
				#$region->farmsteadmap->save_as_ascii("testmap.txt");
				$farmsteadmap=  $region->farmsteadmap;
				
			}
			$mapregistry->add_or_replace_objects($farmsteadmap);
			for my $curAgent ($agentregistry->object_list() ) {
				my %loc2 = %loc;
				$loc2{agent_id} = $curAgent->get_attribute("id");
				$loc2{agent}	= $curAgent;
				do_post_assignment($disinfo, $curAgent, undef, \%loc2);
			}
			
		}	
		elsif ($disinfo->{TYPE} eq "maps_assign_plots") {
			my %loc = %$locvarhash;
			my $smapname = MpmasQLFunctionLanguage->convert($disinfo->{MAP},\%loc);
			
			my $suitabilitymap = $mapregistry->get_object_by_id($smapname);
	  		print STDERR "Unknown MAP: $smapname\n" and exit(1) unless (defined $suitabilitymap);
			$suitabilitymap->reset_nodata_to(-1);
			my $propertymap = $mapregistry->get_object_by_id("PROPERTYMAP");
			my $farmsteadmap= $mapregistry->get_object_by_id("FARMSTEADMAP");
			print STDERR "FARMSTEADMAP undefined\nYou need to distribute farmsteads before you can use maps_assign_plots\n" and exit(1) unless (defined $farmsteadmap);

			my @attr_to_distribute;
			foreach ( @{$disinfo->{SUITABLE}->{lines}}) {
				my ($attr,$suit) =  split /\s*#=\s*/, $_, 2;
				my ($attlist, $attfor) =  MpmasQLFunctionLanguage->translate_and_interprete($attr,\%loc);
				
				foreach my $k (0..$#{$attlist}) { 
						my %hash;
						$hash{field} = $attlist->[$k];
						$hash{content} = $suit;
						$hash{thisfor} = $attfor->[$k];
						push @attr_to_distribute, \%hash;
				} ;
			}
			for my $att (@attr_to_distribute) {
				
				my %agentgroups;
				my %suits;
				#only loop through agents to determine suitability if it is agent-specific 
				if ($att->{content} =~ /#this\s*\(\s*agent/ ) { 
				
					for my $curAgent ($agentregistry->object_list ) {
						my %loc2 = %loc;
						$loc2{agent_id} = $curAgent->get_attribute("id");
						$loc2{agent}	= $curAgent;
						$loc2{attribute} = $att->{thisfor};
						my @suitable = sort {$a <=> $b} MpmasQLFunctionLanguage->convert($att->{content},\%loc2);
						my $suitcode = join ("_", @suitable);
						push @{$agentgroups{$suitcode}}, $curAgent ;
						$suits{$suitcode} = \@suitable;
						#print "Agent $loc{agent_id} , suitcode $suitcode\n" if $verbose;
					}
				}
				else {
					my %loc2 = %loc;
					$loc2{attribute} = $att->{thisfor};
					my @suitable = sort {$a <=> $b} MpmasQLFunctionLanguage->convert($att->{content},\%loc2);
					my $suitcode = join ("_", @suitable);
					$suits{$suitcode} = \@suitable;
					$agentgroups{$suitcode}=[( $agentregistry->object_list() )];
				}
				for my $suitcode (keys %agentgroups) {
					print "Distributing $att->{field}  properties to ". scalar ( @{ $agentgroups{$suitcode} }) ." agents of suitability group $suitcode\n" ;
					my $region = MasMapRegion->create();
					$region->set_farmsteadmap($farmsteadmap);
					$region->set_propertymap($propertymap) if (defined $propertymap);
					$region->add_agents(@{ $agentgroups{$suitcode}});
					$region->set_option("DISTRIBUTE_ORDER", defined $disinfo->{"DISTRIBUTE_ORDER"}? $disinfo->{"DISTRIBUTE_ORDER"}: 0);
					$region->set_option("PLOT_PDF", defined $disinfo->{"PLOT_PDF"} ? $disinfo->{"PLOT_PDF"}: 2 );
					my ($rc, $unfinished_agentlist) = $region->distribute_properties($suitabilitymap, $suits{$suitcode},  $disinfo->{AVGPLOTSIZE}, $att->{field}, $disinfo->{CELLSIZE}, $disinfo->{MAXIT}, $verbose);
					unless ($rc ) {
					    print STDERR "Could not assign all properties to all agents in suitability group $suitcode\n" ;
					    #print "Unfinished Agents\n".join("\n",@$unfinished_agentlist)."\n";
					}
					$region->propertymap->save_as_ascii("testproperty_$att->{field}_$suitcode.txt") if ($verbose > 2);
					$propertymap=  $region->propertymap;	
					$propertymap->set_id("PROPERTYMAP");
				}
				
			}
			$mapregistry->add_or_replace_objects($propertymap);
			for my $curAgent ($agentregistry->object_list() ) {
				my %loc2 = %loc;
				$loc2{agent_id} = $curAgent->get_attribute("id");
				$loc2{agent}	= $curAgent;
				do_post_assignment($disinfo, $curAgent, undef, \%loc2);
			}
		}
		elsif ($disinfo->{TYPE} eq "none") {
			#do nothing
		}
		else {
			print STDERR "Unknown distribution type $disinfo->{TYPE} in STEP $d\n";
			exit(1);
		}
	} continue {
		$total_agent_registry->add_objects($agentregistry->object_list());
		$agentregistry->clear();
		print "Number of agents after STEP $d, GROUP @$clu: ". $total_agent_registry->counter()."\n" if $verbose;
		
		
	}
	for my $curAgent ($total_agent_registry->object_list) {
		my %locvars1;
		$locvars1{agent_id} = $curAgent->id;
		$locvars1{agent} = $curAgent;
		if (exists $disinfo->{DROP_IF}) {
			my $drop = MpmasQLFunctionLanguage->convert( $disinfo->{DROP_IF},\%locvars1);
			if ($drop) {
				$total_agent_registry->delete_objects($curAgent);
				print "Deleted agent ".$curAgent->id.".($drop)\n" if $verbose > 2;
				next;
			}
			
		}
		for my $i (0..$#{$disinfo->{CLEAN}->{lines}}) {
			
			my ($keys, $fors) = MpmasQLFunctionLanguage->translate_and_interprete( $disinfo->{CLEAN}->{lines}->[$i],\%locvars1);
			for my $j (0..$#{$keys}) {		
			    $curAgent->drop_attribute($keys->[$j]);
			}
		}
	}
	print "Number of agents after cleaning in STEP $d:". $total_agent_registry->counter()."\n" if $verbose;
	print "\e[K \r";
	my $fh;
	for my $i (0..$#{$disinfo->{OUTPUT}->{lines}}) {
		my $x = $disinfo->{OUTPUT}->{lines}->[$i];
		$x =~ s/^\s*(\w+)//;
		my $what = $1;
		
		my $fn = "STDOUT";	
		if (not $x =~ /^\s*$/) {
			$fn = MpmasQLFunctionLanguage->convert($x,{});
			open ($fh, ">", $fn) or print STDERR "ERROR when trying to open $fn: $!\n" and exit(1);	
		}
		else {
			open ($fh, ">&STDOUT");	
		}
		if ($what eq "ASSIGNMENTS") {
			print "Writing ASSIGNMENTS to $fn\n";
			for my $ag ($total_agent_registry->object_list()) {
				 my %assets_consolidated;
				map { $assets_consolidated{$_->get_attribute("type") } += $_->get_attribute("size") } $ag->get_attribute("assigned")->object_list() ;
				map { print $fh join("\t", $ag->id, $_, $assets_consolidated{$_} )."\n"	} keys %assets_consolidated;
			}	
		}
		elsif ($what eq "MEMBERS") {
			print "Writing MEMBERS to $fn\n";
			for my $ag ($total_agent_registry->object_list()) {
				 my %members_consolidated;
				map { $members_consolidated{$_->get_attribute("type") } += $_->get_attribute("number") } $ag->get_attribute("memberlist")->object_list() ;
				map { print $fh join("\t", $ag->id, $_, $members_consolidated{$_} )."\n"	} keys %members_consolidated;
			}	
		}
		elsif ($what eq "NOTASSIGNED") {
			print "Writing NOTASSIGNED to $fn\n";
			for my $asset ($notassigned->object_list()) {	
				next unless(ref $asset);
				print $fh $asset->get_attribute("id");
				for my $att ( $asset->attribute_list) {
					next if ($att eq "id");
					print $fh "\t".$att."=".$asset->get_attribute($att);
					
				}	
				print $fh "\n";
			}	
		}
		elsif ($what eq "AGENTS") {
			print "Writing AGENTS to $fn\n";
			for my $agent ($total_agent_registry->object_list()) {
				next unless(ref $agent);
				for my $att ( sort{$a cmp $b} $agent->attribute_list) {
					next if (is_in_list($att,[("id", "assigned")]));
					print $fh join("\t", $agent->get_attribute("id"),$att,$agent->get_attribute($att))."\n";
				}	
				
			}	
		}
		elsif ($what eq "STORE_AGENTS") {
			print "Storing AGENTS to $fn\n";
			$total_agent_registry->save($fn);	
		}
		elsif ($what eq "STORE_NOTASSIGNED") {
			print "Storing NOTASSIGNED to $fn\n";
			$notassigned->save($fn);	
		}
		elsif ($what eq "STORE_SUBAREAS") {
			print "Storing SUBAREAS to $fn\n";
			$subarearegistry->save($fn);	
		}
		elsif ($what =~ /(FARMSTEAD|PROPERTY|SOIL|CLUSTER|POPULATION|NETWORK|SECTOR|CATCHMENT)MAP/) {
			my $which = $1;
			print "Writing ${which}MAP to $fn\n";
			print STDERR "Error: No ${which}MAP has been generated yet!\n" unless (defined $mapregistry->get_object_by_id("${which}MAP"));
			$mapregistry->get_object_by_id("${which}MAP")->save_as_ascii($fn);	
		}

	
		close ($fh);
	}
	($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = localtime(time);
	$mon += 1;
	$year += 1900;
	print "\e[K \rSTEP $d DISTRIBUTION COMPLETE\tTIME: $year $mon $mday , $hour:$min:$sec\n";
}
sub draw_value_from_cdf {
	my $cdf = shift; #hash reference - keys = ubounds - values = size to assign
	my $lb = shift ;
	my $ub = shift ;
	$lb = 0 if (not defined $lb or $lb < 0);
	$ub = 1 if(not defined $ub or $ub > 1);
	my $draw = $lb + rand(1) * ($ub - $lb);
	return $cdf->qf($draw);
	
	
}
sub set_verbosity {
	my $level = shift;
	my $starttag;
	my $endtag;
	if ($level == 0) {
	    $starttag = "";
	    $endtag = "";
	}
	elsif ($level == 1) {
		$starttag = "\e[K \r";	
		$endtag = "\r";
	}	
	elsif ($level >= 2) {
		$starttag = "\e[K \r";	
		$endtag = "\n";
	}	
	return ($starttag, $endtag);
}
sub do_assignment {
	###processing ASSIGN
	my $disinfo = shift;
	my $agent_to_assign_to = shift;
	my $curEndow = shift;
	my %loc2 = %{ shift(@_) };
	my $verbose = shift || 0;
	my ($starttag, $endtag) = set_verbosity($verbose);
	for my $i (0..$#{$disinfo->{ASSIGN}->{lines}}) {
		my ($field, $content) = split /\s*#=\s*/, $disinfo->{ASSIGN}->{lines}->[$i];
		my ($keys, $fors) = MpmasQLFunctionLanguage->convert_to_arrayrefs($field,\%loc2);
		if (defined $keys->[0]) {
			for my $j (0..$#{$keys}) {		
				my %locvars2  = %loc2;
				$locvars2{asset} = $fors->[$j];
				$agent_to_assign_to->add_to_attribute("num_assigned", 1);
				my $newassigned = SimpleObject->create(id => $agent_to_assign_to->get_attribute("num_assigned") );
				$newassigned->set_attribute("type", $keys->[$j]);
				$newassigned->set_attribute("size" , MpmasQLFunctionLanguage->convert($content,\%locvars2) || 0 );
				$agent_to_assign_to->get_attribute("assigned")->add_objects($newassigned);
				
				print $starttag.$agent_to_assign_to->id." receives ". $newassigned->get_attribute("size"). " of ". $keys->[$j].$endtag if $verbose;
			}
		}
		
	}
	
}
sub do_hhmember_assignment {
	###processing MEMBERS
	my $disinfo = shift;
	my $agent_to_assign_to = shift;
	my $curEndow = shift;
	my %loc2 = %{ shift(@_) };
	my $verbose = shift || 0;
	my ($starttag, $endtag) = set_verbosity($verbose);
	if (not defined $agent_to_assign_to->get_attribute("memberlist") ) {
					      $agent_to_assign_to->set_attribute("memberlist", SimpleObjectRegistry->create());
					      $agent_to_assign_to->set_attribute("num_members", 0);
	}
	for my $i (0..$#{$disinfo->{MEMBERS}->{lines}}) {
		my ($field, $content) = split /\s*#=\s*/, $disinfo->{MEMBERS}->{lines}->[$i];
		my ($keys, $fors) = MpmasQLFunctionLanguage->convert_to_arrayrefs($field,\%loc2);
		if (defined $keys->[0]) {
			for my $j (0..$#{$keys}) {		
				my %locvars2  = %loc2;
				$locvars2{member} = $fors->[$j];
				my $x = MpmasQLFunctionLanguage->convert($content,\%locvars2) || 0 ;
				if ($x > 0  ) {
					my $newassigned = SimpleObject->create(id => $agent_to_assign_to->get_attribute("num_members") );
					$newassigned->set_attribute("type", $keys->[$j]);
					$agent_to_assign_to->add_to_attribute("num_members", $x);
					$newassigned->set_attribute("number" , $x );
					$agent_to_assign_to->get_attribute("memberlist")->add_objects($newassigned);
					print $starttag.$agent_to_assign_to->id." receives ". $newassigned->get_attribute("number"). " of ". $keys->[$j].$endtag if $verbose;
				}
				

			}
		}
		
	}
	
}
sub do_post_assignment {
	###processing ON_ASSIGN
	my $disinfo = shift;
	my $agent_to_assign_to = shift;
	my $curEndow = shift;
	my %loc2 = %{ shift(@_) };
	my $verbose = shift || 0;
	my ($starttag, $endtag) = set_verbosity($verbose);
	for my $i (0..$#{$disinfo->{ON_ASSIGN}->{lines}}) {
		
		my ($field, $content) = split /\s*#=\s*/, $disinfo->{ON_ASSIGN}->{lines}->[$i];
		my ($keys, $fors) = MpmasQLFunctionLanguage->convert_to_arrayrefs($field,\%loc2);
		for my $j (0..$#{$keys}) {	
				
			my %locvars2  = %loc2;
			$locvars2{attribute} = $fors->[$j];
			if ($keys->[$j] =~ /(\w+)\.(\w+)/ ) {
				my $e = $1;
				my $f = $2;
				my $ref;
				if ($e eq "observation" and defined $curEndow) {
					$ref = $curEndow;
				}
				elsif ($e eq "agent") {
					$ref = $agent_to_assign_to;
				}
				my $amount = MpmasQLFunctionLanguage->convert($content,\%locvars2) || 0;
				$ref->set_attribute($f, $amount ) if (defined $ref);
				#print "$e $f $amount\n" if $verbose;
			}
		}
		
	}
}
sub evaluate_constraints {
	## Processing constraints
	my $constraints = shift;
	my $fixed_or_dynamic = shift; #arrayref
	my $specificity = shift; #arrayref
	my %loc2 = %{ shift(@_) };
	my $verbose = shift || 0;
	my ($starttag, $endtag) = set_verbosity($verbose);
	foreach my $f (@$fixed_or_dynamic) {
		foreach my $s (@$specificity) {
			for my $i (0..$#{$constraints->{$f}->{$s}}) {
				#print "Evaluating constraint $i for Agent $agent_to_assign_to\n" if $verbose;
				return 0 unless (MpmasQLFunctionLanguage->convert($constraints->{$f}->{$s}->[$i], \%loc2 , 1));				
			}
		}
	}
	return 1;
}
sub divide_constraints {
	my $disinfo = shift;
	my $constraints;
	
	
	for my $j (0..$#{$disinfo->{FIXED_CONSTRAINTS}->{lines}}) {
		my $i = $disinfo->{FIXED_CONSTRAINTS}->{lines}->[$j];
		my $s = 0;
		$s += 1 if ($i =~ /#this\s*\(\s*observation/  );
		$s += 2 if ($i =~ /#this\s*\(\s*agent/ );
		my $t = MpmasQLFunctionLanguage->translate_to_perl($i);
		push @{$constraints->{fixed}->{$s}} , $t;								
	}
	for my $j (0..$#{$disinfo->{DYNAMIC_CONSTRAINTS}->{lines}}) {
		my $i = $disinfo->{DYNAMIC_CONSTRAINTS}->{lines}->[$j];
		my $s = 0;
		$s += 1 if ($i =~ /#this\s*\(\s*observation/  );
		$s += 2 if ($i =~ /#this\s*\(\s*agent/ );
		my $t = MpmasQLFunctionLanguage->translate_to_perl($i);
		push @{$constraints->{dynamic}->{$s}} , $t;				
	}
	return $constraints;
	# fixed /dynamic
	# 1: only observation specific
	# 2: only agent specific
	# 3: agent and observation specific
}

sub fit_into_subarea_draw {
	my ($observations, $likelihoods, $observationregistry, $sum_of_likelihoods) = @_;
	if (not defined $sum_of_likelihoods) {
	      $sum_of_likelihoods = 0;
	      foreach my $curEndow (@$observations) {
		  $sum_of_likelihoods += $likelihoods->{$curEndow->get_attribute("id")};

	      }
	}
	return undef if ($sum_of_likelihoods == 0);
	my %probs;
	foreach my $curEndow (@$observations) {
	    $probs{$curEndow->get_attribute("id")} =  $likelihoods->{$curEndow->get_attribute("id")} / $sum_of_likelihoods;
	}
	my @v = sort {$a <=> $b} keys %probs;
	my @p = map {$probs{$_} } @v;
	my $icdf = EmpiricalDF->new(\@p, \@v, "freq", "step",{discrete => 1});
	# $icdf->qf_plot(undef,0,1,0,20);
	my $obs_id_drawn = draw_value_from_cdf($icdf, 0, 1);
	my $curEndow = $observationregistry->get_object_by_id($obs_id_drawn);
	print "No object of id $obs_id_drawn found\n" and exit(0) unless (defined $curEndow);
	
	return $curEndow;
}

sub prepare_icdf {
	my $curEndow = $_[0];
	my %hash;
	my $df = $curEndow->get_attribute("CDF_ftype") || "cdf";
	my $interpol = $curEndow->get_attribute("CDF_interpolation") || "linear";
	
	my $options = {};
	$options->{discrete} = $curEndow->get_attribute("CDF_discrete") || 0;
	my $icdf;
	
	if ($df =~ /^freq/ or $df eq "pdf") {
			      my %prehash;
			      for my $att ($curEndow->attribute_list) {
				      next unless (($att =~ /^CDF_bin_for_value_([\d.]+)/) or ($att =~ /^CDF_freq_for_value_([\d.]+)/) or ($att =~ /^CDF_prob_for_value_([\d.]+)/));
				      my $value = $1;
				      my $freq = $curEndow->get_attribute($att);
				      print STDERR "Inconsistent frequency $freq found: ".$curEndow->get_attribute("id") ." CDF_bin_for_value_$value\n" and exit(1) if ($freq < 0);
				      $prehash{$value} = 0 unless exists $prehash{$value} ;
				      $prehash{$value} += $freq;
			      }
			      #map {print "$_\t$prehash{$_}\n"} sort {$a <=> $b} keys %prehash;
			      my @v = sort {$a <=> $b} keys %prehash;
			      my @p = map {$prehash{$_} } @v;
			      #security check to ensure function is strictly increasing for Hermite interpolation
			      #Dropping all zero frequencies for the highest numbers.
			      while ( $p[$#p] <= 0) {
				  pop @p;
				  pop @v;
			      }
			      $icdf = EmpiricalDF->new(\@p, \@v, "freq", $interpol,$options);
			      
			     
      
	}
	elsif ($df eq "cdf") { 
	
		my $hundred_found = 0;
		for my $att ($curEndow->attribute_list) {
			next unless ($att =~ /^CDF_bin_for_value_([\d.]+)/);
			my $value = $1;
			my $ubound = $curEndow->get_attribute($att) / 100;
			$hundred_found = 1 if ($ubound == 1);
			print STDERR "Inconsistent upper bound $ubound found: ".$curEndow->get_attribute("id") ." CDF_bin_for_value_$value\n" and exit(1) if ($ubound < 0 or $ubound > 100);
			$hash{$ubound} = $value;
		}
		print STDERR "Highest upper bound for CDF of ".$curEndow->get_attribute("id")." is not 100\n" and exit(1) unless ($hundred_found);
		
		my @p = sort {$a <=> $b} keys %hash;
		my @v = map {$hash{$_} } @p;
		$icdf = EmpiricalDF->new(\@p, \@v, "qf", $interpol, $options);
		
	}
	elsif ($df eq "qf" or $df eq "icdf") { 
	
		my $hundred_found = 0;
		for my $att ($curEndow->attribute_list) {
			next unless ($att =~ /^CDF_value_for_bin_([\d.]+)/);
			my  $ubound = $1 / 100;
			my $value = $curEndow->get_attribute($att) ;
			$hundred_found = 1 if ($ubound == 1);
			print STDERR "Inconsistent upper bound $ubound found: ".$curEndow->get_attribute("id") ." CDF_value_for_bin_$value\n" and exit(1) if ($ubound < 0 or $ubound > 100);
			$hash{$ubound} = $value;
		}
		print STDERR "Highest upper bound for CDF of ".$curEndow->get_attribute("id")." is not 100\n" and exit(1) unless ($hundred_found);
		
		my @p = sort {$a <=> $b} keys %hash;
		my @v = map {$hash{$_} } @p;
		$icdf = EmpiricalDF->new(\@p, \@v, "qf", $interpol, $options);
		
	}
	else {
	    print STDERR "Error: Unknown type of distribution function: $df\n";
	    exit(1);
	}
	$curEndow->set_attribute("CDF_cdf",$icdf);

}
sub predraw_from_uniform {
	my ($potentials,$lbatt, $ubatt) = @_;
	my @plevels;
	if (defined $lbatt and  defined $ubatt) {
	    @plevels = map { 
			    my $lb = $_->get_attribute($lbatt)  || 0; 
			    my $ub = $_->get_attribute($ubatt)  || 100;   
			    $lb/100  + rand(1) * ($ub -$lb) /100;
			  } @$potentials;
	}	
	else {
	    @plevels = map (rand(1), @$potentials) ;
	}
	return \@plevels;
}
