package MpmasMILP;

use strict;
use warnings;
use Exporter;
use Storable qw(dclone);
use Carp qw(croak);
use Data::Dumper;
use MpmasInputfile;
use MpmasInputBlock;
use MpmasInputBlockWithValueReplace;
use IndexedTable;
use SimpleObjectRegistry;
use MpmasMILPActivity;
use MpmasMILPConstraint;
use CtUtilities qw(numbers_robust_to_sorting);
use MpmasQLClassDefinition;
use MpmasQLElementDefinition;
use MpmasQLFieldDefinition;

our @ISA = qw(MpmasInputfile);

use constant  {
  OSL_INFINITY => 10 ** 31
};

sub filename {
	my $self = shift;
	return "MILP".$self->file_suffix if (defined $self->file_suffix);
	return "MILP";
}

sub extension {
	my $self = shift;
	return $self->{extension} if (exists $self->{extension});
	return ".dat";
}

sub set_extension {
	my $self = shift;
	my $str = shift;
	$self->{extension} = $str;
}

sub extension_cm {
	my $self = shift;
	return "$self->{extension_cm}" if (exists $self->{extension_cm});
	return ".txt";
}

sub set_extension_cm {
	my $self = shift;
	my $str = shift;
	$self->{extension_cm} = $str;
}
sub osl_inifinity {
    return OSL_INFINITY;
}
sub prepare_shared {
	my $self = shift;
	
	print "\e[K \rMILP tableau:\n";

	##Determine Order of Constraints
	##Determine Constraints that have been included into the Perennial file and need to go in one block together
	if ($self->config_entry("PERMCROP_MODEL") == 1) {
		for my $i ($self->perennialobjects) {
			for my $j ($i->export_affected_constraints) {
				my $o = $self->constraint($j);
				print STDERR "Constraint $j is referred to in the perennials file, but has not been declared anywhere.\n" and exit (1) unless (defined $o);
				$o->set_attr("type","MASperlab");
			}
		}
	}
	#Sort watersupply constraints correctly:
	


	##Sort Constraints
	my $g = 0;
	my %conrank;
	for (@{$self->config_entry("CONTYPEORDER")}) { $conrank{$_} = ++$g; } ## Define order of constraint types in ini-File

	my $ConstraintTypes;
	for my $element ($self->constraints){
		my $x =$element->attr("type");
		my $y = $element->id;
		if ($x eq "MASwater") {
			$y =~ /(\d+)$/;
			$x = "MASwater1";
			$x = "MASwater2" if ($1 < $self->config_entry("SEASONSTART"));

		}
		print STDERR "Error: constraint type $x not listed in CONTYPEORDER (Constraint $y)\n" and exit(1) unless $conrank{$x};
		print STDERR "Error Constraint $y has been defined twice!\n" and exit(1) if (exists $ConstraintTypes->{$y});
		$ConstraintTypes->{$y} = $x;
	}
	
	my @SortedConstraint = sort {$conrank{$ConstraintTypes->{$a}} <=> $conrank{$ConstraintTypes->{$b}} || numbers_robust_to_sorting($a) cmp numbers_robust_to_sorting($b)  } keys %{$ConstraintTypes} ; ##
	$g = -1;
	for my $i (@SortedConstraint){
		my $x = $self->constraint($i);
		$x->set_lp_row( ++$g);
		print  "Constraint $i placed into row". $x->lp_row()."\n" if $::loudorder;
	
	}
	print "\tConstraints have been sorted\n"if $::loudorder;
	##Determine Order of Activities
	$g = 0;
	
	my %actrank;
	for ( @{$self->config_entry("ACTTYPEORDER")}) { $actrank{$_} = ++$g; } ## Define order of activity types here
	
	
	my $ActivityTypes;
	for my $element ($self->activities){
		#check for existence of type
		my $x = $element->attr("type");
		my $y = $element->id;
		print STDERR "Error Activity $y has been defined twice!\n" and exit(1) if (exists $ActivityTypes->{$y});
		$ActivityTypes->{$y} = $x;
		print STDERR "Error for Activity $y: activity type $x not listed in ACTTYPEORDER\n" and exit(1) unless $actrank{$x};
	      
	}
	my @SortedActivity = sort {$actrank{ $ActivityTypes->{$a}} <=> $actrank{$ActivityTypes->{$b}} || numbers_robust_to_sorting($a) cmp numbers_robust_to_sorting($b) } keys %{$ActivityTypes} ; ##
	$g = -1;
	for my $i (@SortedActivity){
		my $x = $self->activity($i);
		$x->set_lp_column(++$g);
		print  "Activity $i placed into column $g \n" if $::loudorder;
	
	}
	
	my $NumRow = $self->number_of_constraints;
	print "\tActivities have been sorted\n" if $::loudorder;

	my $NumCol = $self->number_of_activities;
	my $IntegAct = 0;
	my $QuadraticAct = 0;
	my ($ZeroCon, $ZeroAct) = (0,0);
	my %exchanges;
	my %number_of_exchanges;
	# count number of Zero Constraints and Matrices
	for my $t ($self->constraints){
		$ZeroCon += 1 if (lc($t->attr("type")) eq "maszero" or lc($t->attr("type")) eq "masforget");	
		#adapt time-variant RHS to chosen start year
		if ($t->has_subelement("exchange")) {
		    for my $sf (sort { numbers_robust_to_sorting($a) cmp numbers_robust_to_sorting($b) } $t->subelement_fields("exchange")) {
			    $sf =~ /(\w+)_(\d+)$/;
			    print STDERR "Error in constraint ".$t->id.", #exchange>: Exogenous change is only possible for the value field of a constraint, not for field '".$1."'\n"  unless ($1 eq "value");
			    #either change value in LP
			    if ($2 <= $self->config_entry("STARTYEAR")) {
				for my $milpid (@{$self->milplist}) {
				    $t->set_attr("value", $t->milpspecific_subelement_field("exchange", $sf, $milpid), $milpid); 
				}
			    }
			    #or include in list of changes of matrix coefficients
			    else {
				my $period = $2 - $self->config_entry("STARTYEAR") + 1;
				for my $milpid (@{$self->milplist}) {
				    $number_of_exchanges{$milpid}++ unless (exists $exchanges{$period}->{$milpid}->{-1}->{$t->lp_row}) ;
				    $exchanges{$period}->{$milpid}->{-1}->{$t->lp_row} = $t->milpspecific_subelement_field("exchange", $sf, $milpid);
				}
				
			    }
		    }
		}
	}
	my @SortedQuadAct;
	for my $t2 (@SortedActivity){
		my $t = $self->activity($t2);
		$ZeroAct +=  1 if (lc($t->attr("type")) eq "maszero" or lc($t->attr("type")) eq "masfuture"  or lc($t->attr("type")) eq "masxc_inc_trans");
		eval {$IntegAct += 1	if ($t->attr("integ") == 1) };
		print $t->attr("name")."\n" if ($t->attr("integ") eq "") ;
	      
		#identify quadratic activities
		if ($t->attr("quadratic") == 1) {
		    $QuadraticAct += 1 ;
		    push @SortedQuadAct, $t2;
		}
		#adapt time-variant coefficients to chosen start year
		if ($t->has_subelement("exchange")) {
		    for my $sf (sort { numbers_robust_to_sorting($a) cmp numbers_robust_to_sorting($b) } $t->subelement_fields("exchange")) {
			    $sf =~ /(\w+)_(\d+)$/;
			    my $what = $1;
			    my $when =$2;
			    if ($what eq "orow") {
				  print STDERR "Error: exogenous change of coefficients of objective row is only possible for activities of type MASzero\nCheck #exchange> orow in activity " .$t->id. "\nUse normal 'orow_<year>' field instead, if this is does not have to be a MASzero activity for other reasons.\n" and exit(1)  unless (lc($t->attr("type")) eq "maszero" );
				  #either change value in LP
				  if ($when <= $self->config_entry("STARTYEAR")) {
				      for my $milpid (@{$self->milplist}) {
					  $t->set_orow($t->milpspecific_subelement_field("exchange", $sf, $milpid), $milpid); 
				      }
				  }
				  #or include in list of changes of matrix coefficients
				  else {
				      my $period = $when - $self->config_entry("STARTYEAR") + 1;
				      for my $milpid (@{$self->milplist}) {
					  $number_of_exchanges{$milpid}++ unless (exists $exchanges{$period}->{$milpid}->{$t->lp_column}->{-1}) ;
					  $exchanges{$period}->{$milpid}->{$t->lp_column}->{-1} = $t->milpspecific_subelement_field("exchange", $sf, $milpid);
				      }
				  }
			    }
			    else {
				  my $c = $self->constraint($what);
				  print STDERR "Error in activity ".$t->id.", #exchange>: Constraint '".$what."' has not been defined anywhere in class file\n"  unless (defined $c);
				  #either change value in LP
				  if ($when <= $self->config_entry("STARTYEAR")) {
				      for my $milpid (@{$self->milplist}) {
					  $t->set_coeff($what, $t->milpspecific_subelement_field("exchange", $sf, $milpid), $milpid); 
				      }
				  }
				  #or include in list of changes of matrix coefficients
				  else {
				      my $period = $when - $self->config_entry("STARTYEAR") + 1;
				      for my $milpid (@{$self->milplist}) {
					  $number_of_exchanges{$milpid}++ unless (exists $exchanges{$period}->{$milpid}->{$t->lp_column}->{$c->lp_row}) ;
					  $exchanges{$period}->{$milpid}->{$t->lp_column}->{$c->lp_row} = $t->milpspecific_subelement_field("exchange", $sf, $milpid);
				      }
				  }
			    }
		    }
		}

	}
	print "\tNumber of Matrix Rows:\t\t $NumRow \n";
	print "\tNumber of Matrix Columns:\t $NumCol \n";
	print "\tNumber of Nonzero Constraints:\t ".($NumRow - $ZeroCon)." \n" unless ($self->config_entry("MPMASCHECKSOFF"));
	print "\tNumber of Nonzero Activities:\t ".($NumCol -$ZeroAct)." \n" unless ($self->config_entry("MPMASCHECKSOFF"));
	print "\tNumber of Integer Activities:\t $IntegAct \n\n";
			
	#look for first labour endowment
	unless ($self->config_entry("MPMASCHECKSOFF"))  {
		for my $i (0..$#SortedConstraint) {
			my $t = $self->constraint($SortedConstraint[$i]);
			if (lc($t->attr("type")) eq "maslab"){
				$self->set_config_entry("FIRSTLAB",  $t->lp_row );
				last;
			}
		}
		print STDERR "\nError: No constraint of type 'maslab' found\n"  and exit (1) unless (defined $self->config_entry("FIRSTLAB"));	
	}

	##Transfer  locals to class globals
	$self->set_sorted_constraints(\@SortedConstraint);
	$self->set_sorted_activities(\@SortedActivity);
	$self->set_sorted_quadratic_activities(\@SortedQuadAct);
	$self->{exchanges} = \%exchanges;
	$self->{number_of_exchanges} = \%number_of_exchanges;
	my $milpinfo = {(
				NumCol => $NumCol,
				NumRow => $NumRow,
				ZeroAct => $ZeroAct,
				ZeroCon => $ZeroCon,		
				IntegAct => $IntegAct,
				QuadraticAct => $QuadraticAct,
				Digs	=> $self->config_entry("DIGS"),
				
			)};
      unless ($self->config_entry("MPMASCHECKSOFF"))  {
	      my $masperlab; 
	      my $maspermac;
	      if ($self->config_entry("PERMCROP_MODEL") == 1) {
		  $masperlab = $self->find_special_constraint("masperlab",0);
		  $maspermac = $masperlab + 1;
	      }
	      else {
		  $masperlab = 	$self->config_entry("FIRSTLAB");
		  $maspermac = $masperlab;
	      }
	      my $masforget = $self->find_special_constraint("masforget",0);
		$milpinfo->{MasPerLab} = $masperlab,
		$milpinfo->{MasPerMac} = $maspermac,
		$milpinfo->{MasForget} = $masforget

      }
	
	$self->set_milpinfo($milpinfo);
	
}
sub write_octave {
	my ($self, $grouping ) = @_;
	
	my $out;		
	
	  my %groupcodes;
	my (@grouplist, @integerlist, @lboundlist, @uboundlist, @orowlist, @coeflist);
	foreach my $act ($self->sorted_activities){
		my $x = $self->activity($act);
		my $col = $x->lp_column;
		my $group = -1;
		$group = $x->get_group($grouping) if (defined $x->get_group($grouping));
		$groupcodes{$group} = -2;
		push @grouplist, $group;


		push @integerlist, $x->attr("integ", $self->argument) ? 73 : 67 ;
		push @lboundlist, $x->attr("lbound", $self->argument);
		push @uboundlist, $x->attr("ubound", $self->argument);
		push @orowlist, (lc($x->attr("type", $self->argument)) eq "masfoodbuy") ? 0 :  $x->orow(undef, $self->argument);
      
		for my $rowelement ($x->coefficients){
			die "Error!!!: Constraint $rowelement of Activity $act is not declared anywhere in class file.\n" unless $self->constraint($rowelement); 
			my $co = $self->constraint($rowelement);
			die "Error!!!: Constraint $rowelement of Activity $act is not declared anywhere in class file.\n" unless defined $co; 
			push @coeflist,  [($co->lp_row + 1, $col +1, $x->coeff($rowelement, $self->argument)  )] if (  $x->coeff($rowelement, $self->argument) != 0) ;
			print STDERR "Error: $act\t $rowelement \t".$co->id."\t".$x->coeff($rowelement, $self->argument)."\n" and exit(1) if ($x->coeff($rowelement, $self->argument) =~ /[^\s\d.+Ee-]/);
		}

	}
	my $c = 0;
	for my $g (sort {$a cmp $b} keys %groupcodes) {
	      if ($g eq "-1") {
		  $groupcodes{$g} = -1;
	      }
	      else {
		  $groupcodes{$g} = ++$c;
	      }
	}
	my @gcodelist = map {$groupcodes{$_}} @grouplist;
	my (@eqtypelist, @rhsvalues);
	my %eqconv = (
	    1 => 85,
	    2 => 76,
	    3 => 83
	);
	foreach my $con ($self->sorted_constraints){
		my $x = $self->constraint($con);
		push @eqtypelist, $eqconv{$x->attr("eqtype")};
		push @rhsvalues,0;
	}
	my $fn = $self->outdir().$self->mname()."_OMILP.dat";
	open($out, ">", $fn ) or die "Can't open $fn: $!";
	print "\e[K \r";
	print "Writing $fn...\r";
	print $out $self->milpinfo("NumRow")."\t".$self->milpinfo("NumCol")."\n";

	print $out join("\t", @gcodelist)."\n";
	print $out join("\t", @integerlist)."\n";
	print $out join("\t", @lboundlist)."\n";
	print $out join("\t", @uboundlist)."\n";
	print $out join("\t", @orowlist)."\n";
	print $out "\n";
	print $out join("\t", @eqtypelist)."\n";
	print $out join("\t", @rhsvalues)."\n";
	print $out "\n";
	#coefficients
	for my $c ( @coeflist) {
	    print $out join("\t", @{$c})."\n";
	}
	#print "..done!\n";
	close ($out);
	return \%groupcodes;
}

sub prepare_special {
	my ($self, $singlemode, $curRHS, $rhshash) = @_;	
	#print "UNDEFINED!\n" unless (defined $rhshash);
	$self->set_extension(".dat");
	$self->set_extension_cm(".txt");
	if ($self->tableau_compact) {
	    $self->set_extension(".cpt");
	}
	
	my $NumRow = $self->milpinfo("NumRow");
	my $NumCol = $self->milpinfo("NumCol");
	my $ZeroAct = $self->milpinfo("ZeroAct");
	my $ZeroCon = $self->milpinfo("ZeroCon");
	my $digs = $self->milpinfo("Digs");

	
	###Consistency check
	my @checkact = $self->sorted_activities();
	die "Number of Activities inconsistent: ". scalar @checkact. " != ".$NumCol."\n" unless (scalar @checkact == $NumCol);

	#Matrix Header
	my $matrixtitle = MpmasInputBlock->new(8,1);
	$matrixtitle->add_comment_row("THE MATRIX FILE");
	$matrixtitle->add_comment_row("");
	$matrixtitle->add_comment_row("1", "PARAMETERS AND COEFFICIENTS"); 
	$matrixtitle->add_comment_row("1.1","Information on the size of the programming matrix","","","","","","","Number of constraints","Number of activities");			
	$matrixtitle->add_row("","","","","","","","","$NumRow","$NumCol");
	$matrixtitle->add_row("");
	unless ($singlemode)  {
		$matrixtitle->add_comment_row("1.2","Other information about the matrix","","","","","","","Number nonzero constraints","Number nonzero activities","Proportion of matrix filled");
		$matrixtitle->add_row("","","","","","","","",$NumRow - $ZeroCon,$NumCol - $ZeroAct, "0.3");	
		$matrixtitle->add_row("");
			
		my  @temp = ("1.3","Constraint Indices","","","","","","","Constraint index of the first labor endowment","Constraint index of the first liquidity endowment");
		if ($self->config_entry("PERMCROP_MODEL") == 1) {
			for my $i ($self->sorted_constraints) {
					my $t = $self->constraint($i);
					push @temp,  $t->attr("name") if (lc($t->attr("type")) eq "masinv");
			}
		}
		if ($self->config_entry("COUNTRY") == 5 ) {
					push @temp, "Cash flow constraint (Coeff.)";
					push @temp, "Cash flow constraint (RHS fix cost)";
					push @temp, "Cash flow constraint (RHS min. consumption)";
		}
		push @temp, "First soil constraint";
		$matrixtitle->add_comment_row(@temp);
		undef @temp;
		@temp[0..7] = "";
		push @temp ,$self->config_entry("FIRSTLAB");
		push @temp, $self->find_special_constraint("masliq",1);		
		if ($self->config_entry("COUNTRY") == 5 ) {
					my $t = -1;
					my $sc = $self->constraint($self->config_entry("CASHFLOW_COEF"));
					if (defined $sc) {
					    $t = $sc->lp_row();
					}
					push @temp, $t ;
					$t = -1;
					$sc = $self->constraint($self->config_entry("CASHFLOW_RHS"));
					if (defined $sc) {
					    $t = $sc->lp_row();
					}
					push @temp, $t;
					$t = -1;
					$sc = $self->constraint($self->config_entry("CASHFLOW_RHS_MINCONS"));
					if (defined $sc) {
					    $t = $sc->lp_row();
					}
					push @temp, $t;
		}
		#  look for constraint indices of investment soil constraints
		
		if ($self->config_entry("PERMCROP_MODEL") == 1) {
			my $csoil = 0;
			for my $i ($self->sorted_constraints) {
					my $t = $self->constraint($i);
					push @temp,  $t->lp_row and $csoil++ if (lc($t->attr("type")) eq "masinv");
			}
			
			print STDERR "Error: Using the permanent crops submodel requires including a investment constraint for each soil type (type: MASinv).\nFound $csoil investment constraints, NSOIL: ".$self->config_entry("NSOIL")."\n" and exit(1) if ($csoil != $self->config_entry("NSOIL"));
		}
			#	look for Constraint index of the first soil endowment"	;
		push @temp, $self->find_special_constraint("massoil", 1);		
		$matrixtitle->add_row(@temp);
		undef @temp;
		
		
		$matrixtitle->add_comment_row("1.4","Indices of Special Constraints","","","","","","", "Index of the average investment cash constraint", "Perennials, index of labor constraint nr. 0\tPerennials, index of machinery constraint nr. 0");
		
		@temp[0..7] = "";
		#look for ongoing liquidity constraint
		push @temp, $self->find_special_constraint("masong",1);
		#	look for Constraint index for the first perennials labour use constraint"	;

		if ($self->config_entry("PERMCROP_MODEL") == 1 ) { 
			push @temp, $self->milpinfo("MasPerLab"); 		
			push @temp, $self->milpinfo("MasPerMac");
		}else{
			push @temp,("");
		}
		$matrixtitle->add_row(@temp);
		undef @temp;
		
		$matrixtitle->add_comment_row("1.5", "Indices of special activities related to cash flows","","","","","","", "Index of the savings activity","Number of short-term credit sources","Index of the first short-term credit activity",	"Number of short-term deposit destinations",	"Index of first short-term deposit activity")	;

		@temp[0..7]  = "";
		if ($self->config_entry("CONSTYPE") == 1) {
			push @temp, $self->find_special_activity("masrcash",1);#remaining cash activity
			#obsolete: push @temp, $self->find_special_activity("masextra",1);# extra household consumption activity
		}
		else {
			push @temp , ("","");
		}
		push @temp, $self->find_special_activities("masscred");	# short-term credit
		push @temp, $self->find_special_activities("masdep",1);	# short-term savings
		$matrixtitle->add_row(@temp);
		undef @temp;
		
		$matrixtitle->add_comment_row("1.6","Indices of special activities related to labor use","","","","","","","Number of hiring in temporary","Index of first activity", "Number of hiring out temporary","Index of first activity", "Number of permanent hiring in","Index of first activity",	"Number of permanent hiring out","Index of first activity");
		@temp[0..7]  = "";		
		push @temp, $self->find_special_activities("maslabtin");# temporary hiring in
		push @temp, $self->find_special_activities("maslabtout");# temporary hiring out
		push @temp, $self->find_special_activities("maslabpin");# permanent hiring in
		push @temp, $self->find_special_activities("maslabpout");# permanent hiring out
		$matrixtitle->add_row(@temp);
		undef @temp;
		
		$matrixtitle->add_comment_row("1.7","Labour Switches","","","","","","", "Number of Labor spans", "Number of Cash spans", "Fixing female labor binary [yes/no]", "Female labor switch: activity index if male and female labor","Female labor switch: activity index if only male labor", "Male labor switch: activity index if male and female labor","Male labor switch: activity index if only male labor");

		@temp[0..7]  = "";
		#Num Labour spans set to one!
		#Note: labor spans are not the same as labor groups, theoretically you could have more than one span per labor group, e.g. for seasons, usually you can however solve this within the MILP (with transfer activities), so for now this has not been implemented in mpmasql
		push @temp, 1;
		my ($count, $first) = $self->find_special_constraints("masliq");		#Num Cash spans
		push @temp, $count;
		# !!CT implement	female/male labour switches !!!
		push @temp, 0;
		$matrixtitle->add_row(@temp);
		undef @temp;
	}	
	$matrixtitle->add_comment_row("1.8","Information on the shape of the programming matrix","","","","","","","Number of matrix blocks","Is the matrix transposed?  (yes=1 / no=0)","Index of  first constraint  in block 0","Index of  first activity  in block 0","Index of  first constraint  in block 1","Index of  first activity  in block 1","Index of  first constraint  in block 2","Index of  first activity  in block 2") ;	
	$matrixtitle->add_row("","","","","","","","",1,0,0,0);
	$matrixtitle->add_row("");
	$matrixtitle->add_row("");
		
	
	$self->add_block($matrixtitle);

	#prepare Objective function; note: in DBG-Modes it is written into the Head of the Tableau
	my $objrow = MpmasInputBlock->new(9,20);
	my @o1; #Scientific Number format
	$#o1 = 8;
	@o1 = ("") x @o1;
	my @o2; #Decimal Number format
	for my $i ($self->sorted_activities) {
		my $fd;
		my $h =	$self->activity($i);
		if (lc($h->attr("type", $self->argument)) eq "masfoodbuy") { #Note: food cunsumption is neutral with respect to income, but price needs to go to market file anyway
		    push @o1,  sprintf("%e", 0);
		    push @o2,  sprintf("%.".$digs."f", 0);
		}
		else {
		    push @o1,  sprintf("%e", $h->orow( undef ,$self->argument));
		    push @o2,  sprintf("%.".$digs."f", $h->orow( undef ,$self->argument));
		}
	}
	
	$objrow->add_comment_row ("2",uc("Objective Function Coefficients"));
	$objrow->add_noncommentonly_row (@o1);
	$objrow->add_commentonly_row("");
	$objrow->add_row("");
	$objrow->add_row("");
	$self->add_block($objrow);		

	##write head of Tableau if in DBG mode
	my @empty;
	$#empty = 101;
	@empty = ("") x @empty;	

	my $matrixtableau = MpmasInputBlock->new(9,30);
	$matrixtableau->add_comment_row("3",uc("Programming Matrix"));
	$matrixtableau->add_comment_row("block1",@empty[0..5], "mpmasql id","","", $self->sorted_activity_attr("id") );
	$matrixtableau->add_comment_row(@empty[0..6], "mpmasql type","","", $self->sorted_activity_attr("type") );
	$matrixtableau->add_comment_row(@empty[0..6], "Name","","", $self->sorted_activity_attr("name") );
	$matrixtableau->add_comment_row(@empty[0..6], "Unit","","", $self->sorted_activity_attr("unit") );
	$matrixtableau->add_comment_row(@empty[0..6], "Solution vector","","",  );
	$matrixtableau->add_comment_row(@empty[0..6], "Objective function","","",@o2);
	$matrixtableau->add_comment_row(@empty[0..6], "Integers? (informational, not marked anymore)","","", $self->sorted_activity_vtype( $self->argument) );	
	$matrixtableau->add_comment_row(@empty[0..6], "Fix in third MILP?","","", $self->sorted_activity_attr("consfix", $self->argument) );	
	$matrixtableau->add_comment_row(@empty[0..6], "Lower bound","","", $self->sorted_activity_attr("lbound", $self->argument) );	
	$matrixtableau->add_comment_row(@empty[0..6], "Upper bound","","", $self->sorted_activity_attr("ubound", $self->argument) );		
	$matrixtableau->add_comment_row("mpmasql id","mpmasql type","Equation type","Range","LHS","Sign","RHS","Name","Unit","Index",$self->sorted_activity_attr("LPColumn") );	





	
	if (not $self->tableau_compact) {

		    my @tableau;
		    my @emptyrow; 
		    $#emptyrow = $NumCol - 1;
		    @emptyrow = (0) x @emptyrow; 
		    for my $j (0..$NumRow-1) {  
			push @tableau, [(@emptyrow)];
		    }
		    #make rows
		    foreach my $act ($self->sorted_activities){
			    my $x = $self->activity($act);
			    next if $x->empty;
			    my $lc = $x->lp_column;
			    for my $rowelement ($x->coefficients){
				    my $co = $self->constraint($rowelement);
				    print STDERR "Error!!!: Constraint $rowelement of activity $act is not declared anywhere in class file.\n" and exit(1) unless (defined $co); 
				    next if $co->empty;
				    my $c = $co->lp_row;
				    print STDERR "Error!!!: Constraint $rowelement of activity $act is not declared anywhere in class file.\n" and exit(1)  unless defined $c;
				    print STDERR "Error!!!: Coefficient of activity $act  for constraint $rowelement is not defined\n" and exit(1)  unless defined ($x->coeff($rowelement, $self->argument) );  
				    $tableau[$c][$lc] =  $x->coeff($rowelement, $self->argument) ;
				    print "Error: $act\t $rowelement \t $c\t $tableau[$c][$lc]\n" if ($tableau[$c][$lc] =~ /[^\s\d.+Ee-]/);
			    }
		    }

		    for my $k (0..$NumRow -1 ) { #Loop over constraints
			    my $i = $self->sorted_constraints($k);
			    my $h =	$self->constraint($i);
			    print STDERR  "Constraint $i has not been defined\n" and exit(1) unless (defined $h);
			    my @temp;

		    #	information block left if in DBG mode
			    
				    my $rhsvalue;			
				    if (defined $h->attr("value",$self->argument)) {
					    $rhsvalue = sprintf("%.".$digs."f", $h->attr("value",$self->argument));
				    }
				    elsif ($singlemode and defined $rhshash && exists $rhshash->{$curRHS}->{$i}) {
					    $rhsvalue = sprintf("%.".$digs."f",$rhshash->{$curRHS}->{$i});
				    }
				    elsif ($singlemode and not defined $rhshash) {
					    $rhsvalue = "&get_value(\$cur, \"$i\")";
				    }
				    else {	
					    $rhsvalue = sprintf("%.".$digs."f",0) ;
				    }
			
			    push @temp, ($i,  $h->attr("type"), $h->attr("eqtype",$self->argument), $h->attr("range",$self->argument), "",$h->attr("showsign",$self->argument),$rhsvalue,$h->attr("name"),$h->attr("unit"),$h->lp_row);
		    #	empty columns for input mode to align with commented mode
			    my @temp_sc = @empty[0..9];


			    $matrixtableau->add_commentonly_row( @temp, map {sprintf("%.".$digs."f",$_)} @{$tableau[$k]} );
			    $matrixtableau->add_noncommentonly_row(@temp_sc, map{sprintf("%e",$_)}  @{$tableau[$k]});

			    
		    }
	  }

	elsif ($self->tableau_compact == 1) {	
		my $matrixcoefficients = MpmasInputBlock->new(9,35);
		my $count_coef = 0;
		foreach my $act ($self->sorted_activities){
			my $x = $self->activity($act);
			next if $x->empty;
			my $col = $x->lp_column;

			for my $rowelement ($x->coefficients){
				my $co = $self->constraint($rowelement);
				print STDERR "Error: Constraint $rowelement of Activity $act is not declared anywhere in class file.\n" and exit(1) unless defined $co; 
				next if $co->empty;
				print STDERR "Error: $act\t $rowelement \t".$co->id."\t".$x->coeff($rowelement, $self->argument)."\n" and exit(1) if ($x->coeff($rowelement, $self->argument) =~ /[^\s\d.+Ee-]/);
				if (  $x->coeff($rowelement, $self->argument) != 0)  {
				      $matrixcoefficients->add_row(@empty[0..6],$co->id, $act, $co->lp_row, $col, sprintf("%e", $x->coeff($rowelement, $self->argument))  ) ;
				      $count_coef++;
				}
				
				
			}

		}
		$matrixcoefficients->add_row_at_beginning(@empty[0..7],"Number of Coefficients:",$count_coef );
		$self->add_block($matrixcoefficients);
	}
### end insertion compact tableau format 120203
	$matrixtableau->add_row("");
	$matrixtableau->add_row("");
	$matrixtableau->add_row("");
	$self->add_block($matrixtableau);
	
	##part below tableau
	unless ($singlemode) {
		
		#Zero Constraints
		my $indepcon = MpmasInputBlock->new(3,40);
		$indepcon->add_row("4","FIXED AND INDEPENDENT CONSTRAINTS");
		$indepcon->add_row("","", "Number of constraints", $ZeroCon);
		$indepcon->add_comment_row("","", "Constraint","Index of constraint",	"Value to be inserted");
		my @s = $self->sorted_constraints;
		for my $i ($#s-$ZeroCon + 1..$#s){									
				my $h	= $self->constraint( $s[$i] );
#				if (lc($h->attr("type")) eq "maszero" or lc($h->attr("type")) eq "masforget") 
				$indepcon->add_row("","",$h->id, $i, $h->attr("value",$self->argument) || 0) 	;
		}
		$indepcon->add_row("");
		$self->add_block($indepcon);
	
		if ($self->config_entry("CONSTYPE") == 1) {
			my $disinvest = MpmasInputBlock->new(3,50);			
			$disinvest->add_row("5","CONSTRAINTS FOR NUTRIENT DEMANDS OF HOUSEHOLD");
#			my ($ndisa, $fdisa) =	$self->find_special_activities("MASdisinvest");
#			my ($ndisc, $fdisc) =	$self->find_special_constraints("MASdisinvest");
			my ($nnutrdemc, $fnutrdemc) = $self->find_special_constraints("MASnutdemand");
#			$disinvest->add_row("","Index of first disinvestment activity","",$fdisa);
#			$disinvest->add_row("","Index of first disinvestment constraint","",$fdisc);
			$disinvest->add_row("", "Index of constr. for hh. energy demand","",$fnutrdemc) ;
			$disinvest->add_empty_rows(2);
			$self->add_block($disinvest);
		}	
		#Special adaptations for harvest LP
		my $fixthird = MpmasInputBlock->new(3,60);			
		$fixthird->add_row("6","FIXING OF COLUMNS IN CONSUMPTION/HARVEST");
		$fixthird->add_row("");	
		if ($self->config_entry("CONSTYPE") == 1 or $self->config_entry("HARVEST_DECISION") == 1 ) {
			for my $i ($self->sorted_activity_attr("consfix",$self->argument) ) {
				$fixthird->add_row("","","",$i);	
			}
		}
		$fixthird->add_row("");	
		$self->add_block($fixthird);
		
		#!!CT: Implement Fine Tuning
		my $finetunetop = MpmasInputBlock->new(3,70);			
		$finetunetop->add_row("7","DIFFERENT PARAMETERS FOR CONSUMPTION/HARVEST");
		$finetunetop->add_row("");
		my $numFineTune = 0;
		
		my $finetune = MpmasInputBlock->new(3,71);			
		$finetune->add_comment_row("","ConstraintID","ActivityID", "Index of constraint", "Index of activity","Coef. in harvest stage");
		for my $i ($self->sorted_activities) {
			my $h = $self->activity($i);
			if ($h->has_subelement("harvest") ) {
			    for my $j ($h->subelement_fields("harvest")) {
				$numFineTune++;
				my $c = $self->constraint($j);
				print STDERR "Error: Unknown constraint referenced in 'harvest' subelement of activity $i\n" and exit(1) if (not defined $c);
				$finetune->add_row("",$j,$i, $c->lp_row(), $h->lp_column,$h->milpspecific_subelement_field("harvest", $j, $self->argument));
			    }
			}
		}


		$finetunetop->add_row("","","Number of coefficients:", $numFineTune);
		$finetune->add_row("");	
		$self->add_block($finetunetop);		
		$self->add_block($finetune);			

		# !! CT: TSPC, do not implement but wait on changes in setup there	
		my $tspc = MpmasInputBlock->new(3,80);			
		$tspc->add_row("8","TSPC INPUT FOR CROPPING ACTIVITIES in NRUs");
		$tspc->add_row("");		
		$tspc->add_row("");			
		$self->add_block($tspc);

	}
	my $typeandrange = MpmasInputBlockWithValueReplace->new(3,90);			
	$typeandrange->add_row("9","TYPE AND RANGE OF CONSTRAINTS");

	if ($singlemode) {
		$typeandrange->add_comment_row("","","","Type","","Lower Bound", "","Upper Bound");
		my $varcon = 0;
		for my $i ($self->sorted_constraints) {
			$varcon = 0;
			my $h = $self->constraint($i);
			my $rhsvalue;		
			if (defined $h->attr("value",$self->argument) && $h->attr("value",$self->argument) != 0) {
				$rhsvalue = sprintf("%.".$digs."e", $h->attr("value",$self->argument));
			}	
			elsif ( defined $rhshash && exists $rhshash->{$curRHS}->{$i}) {
				$rhsvalue = sprintf("%.".$digs."e",$rhshash->{$curRHS}->{$i});
			}
			elsif ( not defined $rhshash) {
				$rhsvalue = "&get_value(\"$i\")";
				$varcon = 1;
			}
			else {	
				$rhsvalue = sprintf("%.".$digs."e",0) ;
			}
	    
			my $eqtype = $h->attr("eqtype",$self->argument);
			my $range = $h->attr("range",$self->argument);
			if ($eqtype == 3 and defined $h->attr("epsilon") and $h->attr("epsilon") != 0) {
			      $eqtype	=	1;
			      $rhsvalue	=	$rhsvalue + $h->attr("epsilon") / 2 unless $varcon;
			      $rhsvalue	=	"$rhsvalue + 0.5 * ". $h->attr("epsilon") if $varcon;
			      $range	=	$h->attr("epsilon");
			}      
			my ($c_lbound, $c_ubound);
			if ($eqtype == 1) {
				$c_ubound = $rhsvalue;
				$c_lbound = - OSL_INFINITY;
				if ($range != -1 * OSL_INFINITY ) {
				      $c_lbound = $rhsvalue + $range unless $varcon; ##range expected to be negative, therefore addition
				      $c_lbound = "$rhsvalue + $range" if $varcon;
				  }
			}
			elsif ($eqtype == 2) {
				$c_lbound = $rhsvalue;
				$c_ubound = OSL_INFINITY;
				if ($range != OSL_INFINITY ) {
				      $c_ubound = $rhsvalue + $range unless $varcon;
				      $c_ubound = "$rhsvalue + $range" if $varcon;
				}
			}
			elsif ($eqtype == 3) {
				$c_ubound = $rhsvalue;
				$c_lbound = 0;
			}
			$typeandrange->add_noncommentonly_row("","","",$eqtype, $c_lbound, $c_ubound);
			
		}
	}
	else {
		$typeandrange->add_comment_row("","","","Type","","Range");
		for my $i ($self->sorted_constraints) {
			my $h = $self->constraint($i);
			$typeandrange->add_noncommentonly_row("","","",$h->attr("eqtype",$self->argument),"",$h->attr("range",$self->argument));
		}

	}
	$typeandrange->add_row("");
	$typeandrange->add_row("");
	$self->add_block($typeandrange);


	my $marked = MpmasInputBlock->new(3,100);	
	if ($singlemode) {
		$marked->add_comment_row("10", "MARKED INTEGERS, LOWER BOUNDS, UPPER BOUNDS ON COLUMNS");
		$marked->add_row("");
		for my $i ($self->sorted_activities) {
			my $h = $self->activity($i);
			my $vtype = 0;
			$vtype = 2 if ($h->attr("quadratic", $self->argument) and not $h->attr("integ", $self->argument) );
			$vtype = 3 if ($h->attr("quadratic", $self->argument) and $h->attr("integ", $self->argument) );
			$marked->add_noncommentonly_row("","","", $vtype   ,$h->attr("lbound",$self->argument), $h->attr("ubound",$self->argument) );#new
		}
	}
	else {		
		$marked->add_comment_row("10","MARKED INTEGERS");
		$marked->add_row("");
		for my $i ($self->sorted_activities) {
			my $h = $self->activity($i);
			my $vtype = 0;
			$vtype = 2 if ($h->attr("quadratic", $self->argument) and not $h->attr("integ", $self->argument) );
			$vtype = 3 if ($h->attr("quadratic", $self->argument) and $h->attr("integ", $self->argument) );
			$marked->add_noncommentonly_row("","","",$vtype); #new
		}
	}
	$marked->add_row("");
	$marked->add_row("");
	$self->add_block($marked);
	
	unless ($singlemode) {
		my $bounds = MpmasInputBlock->new(3,110);	
	
		$bounds->add_comment_row("11", "LOWER BOUNDS ON COLUMNS") ;	
		$bounds->add_row("");
		for my $i ($self->sorted_activity_attr("lbound",$self->argument)) {
			$bounds->add_noncommentonly_row("","","",$i);
		}			
						
		$bounds->add_comment_row("12", "UPPER BOUNDS ON COLUMNS");	
		$bounds->add_row("");
		for my $i ($self->sorted_activity_attr("ubound",$self->argument) ){
			$bounds->add_noncommentonly_row("","","",$i);
		}
		$bounds->add_row("");
		$bounds->add_row("");
		$self->add_block($bounds);
	}
	my $isets = MpmasInputBlock->new(1,130);	
	
	my $curInteg = 0;
	for my $i ($self->sorted_activities) {
		my $h =	$self->activity($i);
		if (defined $h->attr("sos") and $h->attr("sos") ne "-1") {
		    #add activity to SOS
		    my $sos = $self->SOS($h->attr("sos"));
		    print STDERR "Error: Activity ".$h->id." has been included into SOS ".$h->attr("sos").", but ".$h->attr("sos") ." has not been defined anywhere in the .cll file\n" and exit(0) unless (defined $sos);
		    $sos->add_activities($h);
		}
		elsif ($h->attr("integ") == 1 ) {
			$curInteg++;
			$isets->add_row("Integer info Nr. $curInteg","");
			$isets->add_row("Number of variables in set", 1);
			$isets->add_row("Type of set", 4);
			$isets->add_row("Priority", $h->attr("priority",$self->argument) );
			$isets->add_row("Column index", $h->lp_column);
			$isets->add_row("Downside pseudo costs", $h->attr("pseudo_down",$self->argument));
			$isets->add_row("Upside pseudo costs", $h->attr("pseudo_up",$self->argument));
			$isets->add_row("");
		}	

		
	}
	$self->add_block($isets);	
	my $so_sets = MpmasInputBlock->new(1,131);
	my $empty_sos = 0;
	for my $h ($self->SOSs) {
		if ($h->number_of_activities_in_set <= 0) {
		    $empty_sos++;
		    next;
		}
		print STDERR "Error: Invalid type '".$h->attr("type"). "' for SOS ". $h->id."\n" and exit(1) unless ($h->attr("type") > 0 and $h->attr("type") < 4);
		$so_sets->add_row("SOS ".$h->name,"");
		$so_sets->add_row("Number of variables in set", $h->number_of_activities_in_set);
		$so_sets->add_row("Type of set", $h->attr("type") );
		$so_sets->add_row("Priority", $h->attr("priority",$self->argument) );
		my @cindices;
		foreach ($h->activities) { push @cindices, $_->lp_column; }
		$so_sets->add_row("Column indices", @cindices);
		my @cdownpseudo;
		foreach ($h->activities) { push @cdownpseudo, $_->attr("pseudo_down"); }
		$so_sets->add_row("Downside pseudo costs", @cdownpseudo);
		my @cpositions;
		foreach ($h->activities) { push @cpositions, $_->attr("sos_position"); }
		$so_sets->add_row("Reference row", @cpositions);
		$so_sets->add_empty_rows(1);	
	}
	$self->add_block($so_sets);

	my $isetshead = MpmasInputBlock->new(1,129);	
	$isetshead->add_comment_row("13", "INTEGER SETS");
	$isetshead->add_row("");
	$isetshead->add_row("Total number of integer sets", $curInteg );
	$isetshead->add_row("Total number of SOS", $self->get_SOS_registry->counter - $empty_sos);	
	$self->add_block($isetshead);	

	
 	if ( $self->milpinfo("QuadraticAct") > 0) {
	    my $quadratics = MpmasInputBlock->new(2,135);
	    $quadratics->add_comment_row(15, "Quadratic terms");			
	    $quadratics->add_comment_row("", "Number of quadratic variables", $self->milpinfo("QuadraticAct"));
	    $quadratics->add_empty_rows(2);
	    $quadratics->add_row("", "Use quadratic barrier algorithm for continuous quadratic problems", 0) unless ($singlemode);
	    $quadratics->add_comment_row("", "Q matrix for standard form" );
	    $quadratics->add_empty_rows(2);
	    for my $i ($self->sorted_quadratic_activities) {
		my $h = $self->activity($i);	
		my @row = ("","$i");
		for my $j ($self->sorted_quadratic_activities) {
			      my $k = $self->activity($j);	
#			      print "$i; $j; ".$h->subelement_field("quadratic", $j)."\n";
			      print STDERR "Error: Inconsistent element of Q matrix for activities $i and $j:\n $i -> $j = ".$h->subelement_field("quadratic", $j)."\n $j -> $i = ". $k->subelement_field("quadratic", $i). "\n" and exit(1) if (defined $k->subelement_field("quadratic", $i) and defined $h->subelement_field("quadratic", $j) and $k->subelement_field("quadratic", $i) != $h->subelement_field("quadratic", $j) );
			      if (defined $h->subelement_field("quadratic", $j) ) {
				  push @row, $h->subelement_field("quadratic", $j);
			      }
			      elsif (defined $k->subelement_field("quadratic", $i)) {
				  push @row, $k->subelement_field("quadratic", $i);
			      }
			      else {
				  push @row, 0;
			      }
			  
		}
		  $quadratics->add_row(@row);
	    }	
	    $quadratics->add_empty_rows(2);
	    $self->add_block($quadratics);

	}
	unless ($singlemode) {


		my $epsila = MpmasInputBlock->new(2,140);
		$epsila->add_comment_row (17, "APPLY LOWER AND UPPER BOUNDS OF EPSILON TO THE FOLLOWING EE CONSTRAINTS:");			
		$epsila->add_comment_row("", "Note: You can provide individual epsilon for each EE constraint");
		
		my $epsilatab = MpmasInputBlock->new(2,150);	
		my $cc = 0;
		for my $k ($self->sorted_constraints) {
			my $kk = $self->constraint($k);
			if ($kk->attr("eqtype") == 3 and  defined $kk->attr("epsilon")) {
				$epsilatab->add_row("","",$kk->lp_row, sprintf("%e", $kk->attr("epsilon",$self->argument)));
				$cc++;
			}

		}
		$epsila->add_row("","Number of EE contraints", $cc);
		$epsilatab->add_empty_rows(2);
		$self->add_block($epsila);		
		$self->add_block($epsilatab);		

		
		#Shadow price computation
		my $shadow = MpmasInputBlock->new(2,150);
		$shadow->add_comment_row("18","SAVE SHADOW PRICES, ESPECIALLY FOR MIXED INTEGER PROBLEMS:");									
		$shadow->add_comment_row("", "Note: This feature works only for adjacent constraints, for example all soil types. You might also compute all shadow prices for all constraints but this will take time.");
		my $numshadow = 0;
		if (defined $self->config_entry("SHDW_SWITCH") and $self->config_entry("SHDW_SWITCH") == 1 ) {
			$numshadow = defined $self->config_entry("SHDW_NUM") ? $self->config_entry("SHDW_NUM") : 0;
		}
	
		$shadow->add_row("","Number of shadow prices to be computed",$numshadow);
		if ($numshadow > 0) {
			my $lpmode;
			print STDERR "Error: You turned on shadow price calculation (SHDW_SWITCH), but you did not specify SHDW_LPMODE!" and exit (1) unless (defined $self->config_entry("SHDW_LPMODE") );
			if (uc($self->config_entry("SHDW_LPMODE")) =~ /^\s*P/ or $self->config_entry("SHDW_LPMODE") =~ /\s*1\s*/ ) {
				$lpmode = 1;
			}
			elsif (uc($self->config_entry("SHDW_LPMODE")) =~ /^\s*I/ or $self->config_entry("SHDW_LPMODE") =~ /\s*0\s*/ ) {
				$lpmode = 0;
			}
			else {
				print STDERR "Error: Invalid value for SHDW_LPMODE!" and exit (1);
			}
			$shadow->add_row("","LP mode", $lpmode);
		
			print STDERR "Error: You turned on shadow price calculation (SHDW_SWITCH), but you did not specify SHDW_FIRST!" and exit (1) unless (defined $self->config_entry("SHDW_FIRST") );
			my $k  = $self->constraint($self->config_entry("SHDW_FIRST"));
			print STDERR "Error: Constraint ".$self->config_entry("SHDW_FIRST")." used in configuration entry SHDW_FIRST has not been defined!" and exit (1) unless (defined $k );
			my $firstshdwcon = $k->lp_row;
			$shadow->add_row("","First constraint index", $firstshdwcon);				
		}
		$shadow->add_empty_rows(2);
		$self->add_block($shadow);

		my $timevariants = MpmasInputBlock->new(2,160);
		$timevariants->add_comment_row("19","CHANGE COEFFICIENTS IN LP MATRIX DURING SIMULATION:");
		$timevariants->add_comment_row("","Note: This feature allows to change LP coefficients and RHS values  during simulation, e.g. improved technical coefficients starting in period 5");
		$timevariants->add_empty_rows(1);
		$timevariants->add_row("","Number of matrix elements to be changed", $self->{number_of_exchanges}->{$self->argument} || 0);	
		
		$timevariants->add_comment_row("","", "period","LP column","LP row","value");
		for my $period (sort {$a <=> $b} keys %{$self->{exchanges}}) {
		    for my $col (sort {$a <=> $b} keys %{$self->{exchanges}->{$period}->{$self->argument}}) {
			for my $row (sort {$a <=> $b} keys %{$self->{exchanges}->{$period}->{$self->argument}->{$col}}) {
			    $timevariants->add_row("","",$period, $col, $row, $self->{exchanges}->{$period}->{$self->argument}->{$col}->{$row} );
			}
		    }
		}
		$self->add_block($timevariants);
	}		
}

sub recreate_from_file {
	my ($class, $matrixfile,$mxmode, $actdesname, $condesname, $mpxformat, $mpmas_subformat, $tableau_compact_in) = @_;

	my $activities = IndexedTable->create_from_ascii($actdesname, 1, "ActivityInfo", '\t',1);
	my $constraints = IndexedTable->create_from_ascii($condesname, 1, "ConstraintInfo", '\t',1);
	my $NewConstraintRegistry = SimpleObjectRegistry->create();
	my $NewActivityRegistry = SimpleObjectRegistry->create();
	my $NewSOSRegistry = SimpleObjectRegistry->create();
	MpmasMILPActivity->switch_off_defaults;
	MpmasMILPConstraint->switch_off_defaults;
	MpmasMILPSOS->switch_off_defaults;
	my @quadratics;
	for my $act ($activities->query([],0) ) {
		my %fields;
		$fields{LPColumn} = $act;
		$fields{name} = $activities->query([$act],3);
		$fields{integ} = 0; #set default here
		$fields{type} = $activities->query([$act],5);
		$fields{account} = $activities->query([$act],6);
		$fields{quadratic} = 0;
		my $mn = $activities->query([$act],2);
		my $new = MpmasMILPActivity->create($mn, \%fields, undef, undef ,0);
		$new->set_lp_column($act);
		$NewActivityRegistry->add_objects($new);
	}
	for my $con ($constraints->query([],0) ) {
		my %fields;
		$fields{LPRow} = $con;
		$fields{name} = $constraints->query([$con],3);
		$fields{type} = $constraints->query([$con],5);
		my $mn = $constraints->query([$con],2);
		my $new = MpmasMILPConstraint->create($mn, \%fields,undef, undef ,0);
		$new->set_lp_row($con);
		$NewConstraintRegistry->add_objects($new);
	
	}
	print "Reading matrix from $matrixfile...\n";
	open (my $fn, "<", $matrixfile) or die "Error when opening $matrixfile: $!\n";
	my $part = "pre";

	my ($NumRow, $NumCol, $ninteg, $nsos, %integhash, $percfilled, $NumZeroRow, $NumFineTune, $NumCoef);
	my $cc = 0;
	my $ci = 0;
	LINES: while (<$fn>  ) {	
		chomp;
		my $line = $_;
		$line =~ s/^\t*//;
		
		if ($part eq "pre") {
			next LINES if ($line =~ /^\s*$/);
			next LINES if ($line =~ /^\s*NROWS/ or $line =~ /^\s*NumRows/);
			if ($line =~ /^\s*\d*/ ) {
				$line =~ s/^\s*//;
				$line =~ s/\s*$//;
				($NumRow, $NumCol, $percfilled) = split /\s/, $line;
				print STDERR "Number of Activities in ActivityInfo ".$NewActivityRegistry->counter." does not fit Number of Activities in Matrix File ($NumCol)!\n" and exit(1) unless ($NumCol == $NewActivityRegistry->counter);
				print STDERR "Number of Constraints in ConstraintInfo".$NewConstraintRegistry->counter." does not fit Number of Constraints in Matrix File ($NumRow)!\n" and exit(1) unless ($NumRow == $NewConstraintRegistry->counter);
				$part = "shape";
				$part = "mpmas_info_block" if ($mxmode eq "Mpmas");
				
			}
			else {
				print STDERR "File does not fit specified format ($mxmode)\n" and exit(1);
			}
			
		}
		elsif ($part eq "mpmas_info_block") {
				if ($line =~ /^\s*\d+/ ) {
				    $cc++;
				}
				if ($cc == 6) {
				    $cc = 0;
				    $part = "shape"
				}
		}
		elsif ($part eq "shape") {
			#can for now only handle one block
			next LINES if ($line =~ /^\s*$/);
			next LINES if ($line =~ /^\s*NumBlx/);
			if ($line =~ /^\s*\d/ and $cc == 0) {
				$line =~ s/^\s*//;
				$line =~ s/\s*$//;
				my ($blocks, $trans) = split /\s+/, $line;
				print STDERR "Blocked matrices and transposed matrices are not yet supported\nRead blocks: $blocks, transposed: $trans\n"  and exit(1) unless ($blocks == 1 and $trans == 0);
				$cc++;
				if ($mxmode eq "Single" or $mxmode eq "Mpmas") {
					$part = "orow";
					$cc= 0;
				}
			}
			elsif ($line =~ /^\s*\d*/ and $cc == 1) {
				$cc = 0;
				$part = "orow";
			}
		}
		elsif ($part eq "orow") {
			next LINES if ($line =~ /^\s*$/);
			next LINES if ($line =~ /^\s*Matrix Block/);
			next LINES if ($line =~ /^\s*Objective Row/);
			if ($line =~ /^\s*(-?\d)/) {
				my @orow = split /\t+/, $line, $NumCol;
				for my $i (0..$#orow) {
					my $actcode = 	$activities->query([$i], 2);
					my $act = $NewActivityRegistry->get_object_by_id($actcode);
					$act->set_attr("orow", $orow[$i]);
				}
				
				$part = "tableau";
			}
		}
		elsif ($part eq "tableau" and not $mpxformat and not $tableau_compact_in) {
		     
		     
			if ($line =~ /^\s*(-?\d)/ and $cc < $NumRow) {
				my @row = split /\t+/, $line, $NumCol;
				my $concode = 	$constraints->query([$cc], 2);
				for my $i (0..$#row) {
					my $actcode = 	$activities->query([$i], 2);
					my $act = $NewActivityRegistry->get_object_by_id($actcode);
					$act->set_coeff($concode, $row[$i]) if ($row[$i] != 0);
				}				
				$cc++;	
			}
			if ($cc >= $NumRow) {
				$part = "cons";
				$part = "nonzerocons" if ($mxmode eq "Mpmas");
				$cc = 0;
			}
		      
		}
		elsif ($part eq "tableau" and $mpxformat) {
			
		        if ($line =~ /^\s*EquaRHS/) {
			      $part = "cons";
			      $cc = 0;
			      next LINES;
			}
			elsif ($line =~ /^\s*(-?\d)/) {
			      my ( $row, $col, $coef) = split /\s+/, $line;
			      #print join("/", $row, $col, $coef)."\n";
			      my $actcode = 	$activities->query([$col], 2);
			      print STDERR "Activity No. $col not found!\n" and exit(0) unless (defined $actcode);
			      my $concode = 	$constraints->query([$row], 2);
			      print STDERR "Constraint No. $row not found!\n" and exit(0) unless (defined $concode);
			      my $act = $NewActivityRegistry->get_object_by_id($actcode);
			      $act->set_coeff($concode, $coef);
			}
		}
	      elsif ($part eq "tableau" and $tableau_compact_in) {
			if (not defined $NumCoef and $line =~ /^\s*(\d+)/) {
			      $NumCoef = $1;
			}
			elsif ($line =~ /^\s*(-?\d)/) {
			      my ( $row, $col, $coef) = split /\s+/, $line;
			      #print join("/", $row, $col, $coef)."\n";
			      my $actcode = 	$activities->query([$col], 2);
			      print STDERR "Activity No. $col not found!\n" and exit(0) unless (defined $actcode);
			      my $concode = 	$constraints->query([$row], 2);
			      print STDERR "Constraint No. $row not found!\n" and exit(0) unless (defined $concode);
			      my $act = $NewActivityRegistry->get_object_by_id($actcode);
			      $act->set_coeff($concode, $coef);
			      $cc++;
			}
			if (defined $NumCoef and $cc >= $NumCoef) {
			        $part = "cons";
				$part = "nonzerocons" if ($mxmode eq "Mpmas");
				$cc = 0;
			}
		}
		elsif ($part eq "nonzerocons") {
			      next LINES if ($line =~ /^\s*$/);
			      if ($line =~ /^\s*(\d+)/ and not defined $NumZeroRow) {
				    $NumZeroRow = $1;
			      }
			      elsif ($line =~ /^\s*\d+/ and $cc < $NumZeroRow) {
				  my @row = split /\t+/, $line;
				  my $concode =  $constraints->query([$row[0]], 2);
				  my $con = $NewConstraintRegistry->get_object_by_id($concode);
				  $con->set_attr("value", $row[1]);
				  $cc++;
			      }
			      if ($NumZeroRow == 0 or $cc >= $NumZeroRow) {
				  $cc = 0;
				  $part = "fine_tune";
				  $part = "disinvest_info" if ($mpmas_subformat == 2);
				  $part = "consfix_info" if ($mpmas_subformat == 1);
			      }
		}
		elsif ($part eq "disinvest_info") {
			      next LINES if ($line =~ /^\s*$/);
			      if ($line =~ /^\s*(\d+)/) {
				  $cc++;
			      }
			      if ($cc == 1) {
				  $cc = 0;
				  $part = "consfix_info";
			      }		
		}
		elsif ($part eq "consfix_info") {
			      next LINES if ($line =~ /^\s*$/);
			      if ($line =~ /^\s*(\d+)/ and $cc < $NumCol) {
				      my $actcode = $activities->query([$cc], 2);
				      my $act = $NewActivityRegistry->get_object_by_id($actcode);
				      $act->set_attr("consfix", $1);
				      $cc++;
			      }
			      if ($cc >= $NumCol) {
				      $part = "fine_tune";
				      $cc = 0;
			      }
		}
		elsif ($part eq "fine_tune") {
			      next LINES if ($line =~ /^\s*$/);
			      if ($line =~ /^\s*(\d+)/ and not defined $NumFineTune) {
				      $NumFineTune = $1;
			      }
			      elsif ($line =~ /^\s*\d+/ and $cc < $NumFineTune) {
				  $cc++;
			      }
			      if ($NumFineTune == 0 or $cc >= $NumFineTune) {
				  $cc = 0;
				  $part = "cons";
			      }
		}
		elsif ($part eq "cons") {
			next LINES if ($line =~ /^\s*$/);
			next LINES if ($line =~ /^\s*EquaRHS$/);
			
			if ($line =~ /^\s*(-?\d)/ and $cc < $NumRow) {
				my @row = split /\t+/, $line;
				my $concode = 	$constraints->query([$cc], 2);
				my $con = $NewConstraintRegistry->get_object_by_id($concode);
				$con->set_attr("eqtype", $row[0]);
				if ($row[0] == 1 ) {
					if ($mxmode eq "Failed" or $mxmode eq "Single"){
					      $con->set_attr("range", ($row[1] == -1 * OSL_INFINITY) ? $row[1] : $row[1] - $row[2]  );
					      $con->set_attr("value", $row[2]) 	
					}
					elsif ($mxmode eq "Mpmas") {
						$con->set_attr("range", $row[1]);
					}
					$con->set_attr("showsign", "'<=");
				      
				}
				elsif ($row[0] == 2) {
					if ($mxmode eq "Failed" or $mxmode eq "Single") {
						$con->set_attr("value", $row[1]);
						$con->set_attr("range", ($row[2] == OSL_INFINITY) ? $row[2] : $row[2] - $row[1]);
					}
					elsif ($mxmode eq "Mpmas") {
						$con->set_attr("range", $row[1]);
					}
					$con->set_attr("showsign", "'>=");
				}
				elsif ($row[0] == 3 ) {
					if ($mxmode eq "Failed" or $mxmode eq "Single") {
						$con->set_attr("range", $row[1]);
						$con->set_attr("value", $row[2])
					}
					elsif ($mxmode eq "Mpmas") {
						$con->set_attr("range", $row[1]);
					}	
					$con->set_attr("showsign", "'==");
				}
				$con->set_attr("value", 0) unless (defined $con->attr("value"));
				$cc++;
			}
			if ($cc >= $NumRow) {
				$part = "act" if ($mxmode eq "Failed" or $mxmode eq "Single");
				$part = "marked" if ($mxmode eq "Mpmas");
				$cc = 0;
			}
		}
		elsif ($part eq "act") { #only for Failed and Single
			next LINES if ($line =~ /^\s*$/);
			next LINES if ($line =~ /^\s*TypeVar$/);
			
			if ($line =~ /^\s*(-?\d)/ and $cc < $NumCol) {
				my @row = split /\t+/, $line;
				my $actcode = 	$activities->query([$cc], 2);
				my $act = $NewActivityRegistry->get_object_by_id($actcode);
				if ($row[0] >= 2 ) {
				    $act->set_attr("quadratic", 1);
				    push  @quadratics , $act->id;
				}
				$act->set_attr("lbound", $row[1]);
				$act->set_attr("ubound", $row[2]);
				$cc++;
			}
			if ($cc >= $NumCol) {
				$part = "ninteg";
				$cc = 0;
			}
		}
		elsif ($part eq "marked") { #only for Mpmas
			next LINES if ($line =~ /^\s*$/);		
			if ($line =~ /^\s*(-?\d)/ and $cc < $NumCol) {
			      if ($1 >= 2 ) {
				    my $actcode = 	$activities->query([$cc], 2);
				    my $act = $NewActivityRegistry->get_object_by_id($actcode);
				    $act->set_attr("quadratic", 1);
				    push  @quadratics , $act->id;
			      }
			      $cc++;
			}
			if ($cc >= $NumCol) {
				$part = "lbound";
				$cc = 0;
			}
		}
		elsif ($part eq "lbound") { #only for Mpmas
			next LINES if ($line =~ /^\s*$/);
			if ($line =~ /^\s*(-?\d)/ and $cc < $NumCol) {
				my $actcode = 	$activities->query([$cc], 2);
				my $act = $NewActivityRegistry->get_object_by_id($actcode);
				$line =~ s/\s+$//;
				$act->set_attr("lbound", $line);
				$cc++;
			}
			if ($cc >= $NumCol) {
				$part = "ubound";
				$cc = 0;
			}
		}
		elsif ($part eq "ubound") { #only for Mpmas
			next LINES if ($line =~ /^\s*$/);
			if ($line =~ /^\s*(-?\d)/ and $cc < $NumCol) {
				my $actcode = 	$activities->query([$cc], 2);
				my $act = $NewActivityRegistry->get_object_by_id($actcode);
				$line =~ s/\s+$//;
				$act->set_attr("ubound", $line);
				$cc++;
			}
			if ($cc >= $NumCol) {
				$part = "ninteg";
				$cc = 0;
			}
		}
		elsif ($part eq "ninteg") {
			next LINES if ($line =~ /^\s*$/);
			next LINES if ($line =~ /^\s*NumInt$/);
			if ($line =~ /\s*(\d+)/) {
				$ninteg = $1;
				$part = "nsos";
			}
		}
		elsif ($part eq "nsos") {
			next LINES if ($line =~ /^\s*$/);
			next LINES if ($line =~ /^\s*NumSOS$/);
			if ($line =~ /\s*(\d+)/) {
				$nsos = $1;
				$part = "end";
				$part = "Q" if (scalar @quadratics > 0);
				$part = "isets" if ($ninteg > 0);
			}
		}
		elsif ($part eq "isets") {
			next LINES if ($line =~ /^\s*$/);
			if ($line =~ /^\s*(-?\d+\.?\d*)/) {
				if ($cc <= 1) {
					#discard number of activities (asummed to be 1) and type of set (4)
					$cc++;
				}
				elsif ($cc == 2) {
					$integhash{priority} = $1;
					$cc++;
				}
				elsif ($cc == 3) {
					$integhash{act} = $1;
					$cc++;
				}
				elsif ($cc == 4) {
					$integhash{pseudo_down} = $1;
					$cc++;
				}
				elsif ($cc == 5) {
					$integhash{pseudo_up} = $1;
					
					my $actcode = 	$activities->query([ $integhash{act} ], 2);
					my $act = $NewActivityRegistry->get_object_by_id($actcode);
					$act->set_attr("integ", 1);
					$act->set_attr("priority", $integhash{priority});
					$act->set_attr("pseudo_down", $integhash{pseudo_down});
					$act->set_attr("pseudo_up", $integhash{pseudo_up});
					$cc = 0;
					$ci++;
				}

			}
			if ($ci >= $ninteg) {
			      $ci = 0;
			      $part = "end";
			      $part = "Q" if (scalar @quadratics > 0);
			      $part = "sos" if ($nsos > 0);
			      #$part = "eqrelax" if ($mxmode eq "Mpmas");
			}
		}
		elsif ($part eq "sos") {
			next LINES if ($line =~ /^\s*$/);
			if ($line =~ /^\s*(-?\d+)/) {
				if ($cc == 0) {
					$integhash{num_act} = $1;
					$cc++;
				}
				elsif ($cc == 1) {
					$integhash{type} = $1;
					$cc++;
				}
				elsif ($cc == 2) {
					$integhash{priority} = $1;
					$cc++;
				}
				elsif ($cc == 3) {
					$integhash{act} = [ split /\t/, $line ];
					$cc++;
				}
				elsif ($cc == 4) {
					$integhash{pseudo_down} =  [ split /\t/, $line ];
					$cc++;
				}
				elsif ($cc == 5) {
					$integhash{refrow} = [ split /\t/, $line ];
					my $sos =  MpmasMILPSOS->create("sos_$ci",{},{},{}, 0 );
    
					$sos->set_attr("name","sos_$ci");
					$sos->set_attr("type", $integhash{type});
					$sos->set_attr("priority",$integhash{priority}); 
					
					$NewSOSRegistry->add_objects($sos);
					for my $ai ( 0..$#{$integhash{act}} ) { 
						  my $actcode = $activities->query([ $integhash{act}->[$ai] ], 2);
						  my $act = $NewActivityRegistry->get_object_by_id($actcode);
						  $act->set_attr("integ", 1);
						  $act->set_attr("sos", "sos_$ci");
						  $act->set_attr("sos_position", $integhash{refrow}->[$ai]);
						  $act->set_attr("pseudo_down",$integhash{pseudo_down}->[$ai]);
					}
					
					$cc = 0;
					$ci++;
				}

			}
			if ($ci >= $nsos) {
			      $ci = 0;
			      $part = "end";
			      $part = "Q" if (scalar @quadratics > 0);
			      #$part = "eqrelax" if ($mxmode eq "Mpmas");
			}
		}
		elsif ($part eq "Q") {
		  next LINES if ($line =~ /^\s*$/);
		  if ($cc == scalar(@quadratics)) {
		      $part = "end";
		      $cc=0;
		      next;
		   }
		  elsif ($line =~/Number of quadratic variables:\s+(\d+)/) {
		      print STDERR "Warning: number of quadratic variables indicated, does not correspond to number of activities marked as quadratic!" if (scalar(@quadratics) != $1);
		  }
		  # Line "Q matrix in standard form" simply ignored
		  if ($line =~ /^\s*(-?\d+)/) {
		      $line =~ s/^\s*//;
		      my $actcode = $quadratics[$cc];
		      my $act = $NewActivityRegistry->get_object_by_id($actcode);
		      print STDERR "Error: Unknown qudaratic activity $actcode\n" and exit(1) unless (defined $act);
		      my @row = split /\s+/, $line;
		      for my $oth (@quadratics ) {
			my $v = shift @row;
			$act->add_subelement_field("quadratic", $oth, $v);
			#print $act->id."; ". $oth ."; $v\n";
		      }
		      $cc++;
		  }
		}
		if ($part eq "end") {
		    last;
		}
	}
	close $fn;
	print "...done\n";
	print "Initialize matrix structure in memory...\n";
	my @SortedAO = sort {$a->lp_column <=> $b->lp_column} $NewActivityRegistry->object_list;
	my @SortedCO = sort {$a->lp_row <=> $b->lp_row} $NewConstraintRegistry->object_list;
	my @SortedActivity;   
	foreach my $i(@SortedAO) {push @SortedActivity , $i->id  };
	my @SortedConstraint;   
	foreach my $i(@SortedCO) {push @SortedConstraint , $i->id  } ;

	my $self = eval ($class."->new(10,1)");
	$self->set_SOS_registry( $NewSOSRegistry);
	$self->set_constraint_registry( $NewConstraintRegistry);
	$self->set_activity_registry($NewActivityRegistry);
	$self->set_sorted_constraints(\@SortedConstraint);
	$self->set_sorted_activities(\@SortedActivity);
	$self->set_sorted_quadratic_activities(\@quadratics);

	my $milpinfo = {(
				NumCol => $NumCol,
				NumRow => $NumRow,
				ZeroAct => 0,
				ZeroCon => $NumZeroRow || 0,		
				IntegAct => $ninteg,
				SOS	=> $nsos,
				Digs => 4,
				QuadraticAct => scalar(@quadratics)
			)};
	$self->set_milpinfo($milpinfo);

	print "...done\n";
	return $self;
}

sub calculate_lhs_of_constraint {
			my $self = shift;
			my $k = shift; #Constraint
			$k = $self->constraint($k) if (not ref $k);
			my $attribute = shift;
			$attribute = "solution" unless (defined $attribute);
			my $sum = 0;
			for my $i (0..$self->milpinfo("NumCol")-1) {
				my $c = $self->activity_at_col($i);
				$sum += $c->coeff($k->id) * $c->attr($attribute) if (defined $c->coeff($k->id));
			}
			return $sum;
}
sub calculate_linear_part_of_objval {
			my $self = shift;
			my $sum = 0;
			for my $i (0..$self->milpinfo("NumCol")-1) {
				my $c = $self->activity_at_col($i);
				$sum += $c->orow() * $c->attr("solution") ;
			}
			return $sum;
}
sub import_solution_vector {
	my ($self,  $solutionfile, $ra) = @_;
	open (my $fh, "<", $solutionfile) or croak "Error when trying to open $solutionfile: $!";
	#print "Importing Solution Vector from $solutionfile\n";
	my $cc = 0;
	my $t = 0;
	while (<$fh>) {
		chomp;
		my $line = $_;

		next if ($line =~ /^\s*$/); #skip empty lines
		if ($line =~ /\s*Target value:\s*?((\s|-)[-\d.E+e]+)/) {
			$t = $1;
			$self->set_milpinfo_entry("TargetValue",$t);
		}
		elsif ($line =~/^\s*?((\s|-)[-\d.E+e]+)/) {
			my $h = $self->activity_at_col($cc);
			$h->set_attr("solution", $1);
			$cc++;
		}
	}
	close $fh;
	if ($self->milpinfo("QuadraticAct") > 0 ) {
		$t = $t * (-1);
		$self->set_milpinfo_entry("TargetValue",$t);
		my $l = $self->calculate_linear_part_of_objval();
		my $q = $t - $l;
		$self->set_milpinfo_entry("ObjectiveValueLin", $l);
		$self->set_milpinfo_entry("ObjectiveValueQuad", $q);
		if (defined $ra) {
	      		$self->set_milpinfo_entry("VarianceRiskModel", $q * (-2) / $ra);
			$self->set_milpinfo_entry("RiskAversionCoef", $ra);
		}
	}
	return 1;
}
sub write {
	
	my ($self, $mode,$cur, $valuetab, $singlemode) = @_;
	my $dir = $self->outdir();
	if ($singlemode) {
		$self->set_extension("_".$cur.".mtx");
		$self->set_extension_cm("_".$cur.".txt");
	} 
	$self->SUPER::write($mode,$cur,$valuetab);
}
sub adhoc_create {
	my $class = shift;
	my $NewConstraintRegistry = SimpleObjectRegistry->create();
	my $NewActivityRegistry = SimpleObjectRegistry->create();
	
	

	my $self = eval ($class."->new(10,1)");
	MpmasInputfile->set_global_references("adhoc", "/var/tmp/", "/var/tmp/", undef, undef, undef, undef, undef, undef,undef);  
	$self->set_constraint_registry( $NewConstraintRegistry);
	$self->set_activity_registry($NewActivityRegistry);

	my $defaults = MpmasQLClassDefinition->create("DEFAULT");
	my $default_activity = MpmasQLElementDefinition->create("activity", "activity");
	$default_activity->add_field(MpmasQLFieldDefinition->create(id => "name", fieldname => "name",  content  => "N/A", type    => "scalar"));
	$default_activity->add_field(MpmasQLFieldDefinition->create(id => "unit", fieldname => "unit",  content  => "N/A", type    => "scalar"));
	$default_activity->add_field(MpmasQLFieldDefinition->create(id => "integ", fieldname => "integ",  content  => "0", type    => "scalar"));
	$default_activity->add_field(MpmasQLFieldDefinition->create(id => "lbound", fieldname => "lbound",  content  => "0", type    => "scalar"));
	$default_activity->add_field(MpmasQLFieldDefinition->create(id => "ubound", fieldname => "ubound",  content  => "10**31", type    => "scalar"));
	$default_activity->add_field(MpmasQLFieldDefinition->create(id => "orow", fieldname => "orow",  content  => "0", type    => "scalar"));
	$default_activity->add_field(MpmasQLFieldDefinition->create(id => "priority", fieldname => "priority",  content  => "1000", type    => "scalar"));
	$default_activity->add_field(MpmasQLFieldDefinition->create(id => "pseudo_down", fieldname => "pseudo_down",  content  => "0", type    => "scalar"));
	$default_activity->add_field(MpmasQLFieldDefinition->create(id => "pseudo_up", fieldname => "pseudo_up",  content  => "0", type    => "scalar"));
	$default_activity->add_field(MpmasQLFieldDefinition->create(id => "type", fieldname => "type",  content  => "default", type    => "scalar"));
	$default_activity->add_field(MpmasQLFieldDefinition->create(id => "xnproduct", fieldname => "xnproduct",  content  => "-1", type    => "scalar"));
	$default_activity->add_field(MpmasQLFieldDefinition->create(id => "xnlink", fieldname => "xnlink",  content  => "-1", type    => "scalar"));
	$default_activity->add_field(MpmasQLFieldDefinition->create(id => "consfix", fieldname => "consfix",  content  => "0", type    => "scalar"));
	$default_activity->add_field(MpmasQLFieldDefinition->create(id => "landfix", fieldname => "landfix",  content  => "0", type    => "scalar"));

	$defaults->add_elements($default_activity );
	my $default_constraint = MpmasQLElementDefinition->create("constraint", "constraint");
	$default_constraint->add_field(MpmasQLFieldDefinition->create(id => "name", fieldname => "name",  content  => "N/A", type    => "scalar"));
	$default_constraint->add_field(MpmasQLFieldDefinition->create(id => "unit", fieldname => "unit",  content  => "N/A", type    => "scalar"));
	$default_constraint->add_field(MpmasQLFieldDefinition->create(id => "eqtype", fieldname => "eqtype",  content  => 1, type    => "scalar"));
	$default_constraint->add_field(MpmasQLFieldDefinition->create(id => "value", fieldname => "value",  content  => 0, type    => "scalar"));
	$default_constraint->add_field(MpmasQLFieldDefinition->create(id => "type", fieldname => "type",  content  => "default", type    => "scalar"));
	$default_constraint->add_field(MpmasQLFieldDefinition->create(id => "xnproduct", fieldname => "xnproduct",  content  => "-1", type    => "scalar"));
	$defaults->add_elements($default_constraint );
	MpmasQLClassDefinition->set_defaults($defaults);
	$self->{adhoc_solve_rm} = "No solution attempt yet!";
	return $self;
}
sub adhoc_add_constraint {
	my ($self, $id, $fields) = @_;
	my $new = MpmasMILPConstraint->create($id, $fields,undef, undef ,0);
	$new->set_lp_row($self->get_constraint_registry->counter );
	$self->add_constraints($new);
	return $new;
}
sub adhoc_add_activity {
	my ($self, $id, $fields) = @_;
	my $new = MpmasMILPActivity->create($id, $fields,undef, undef ,0);
	$new->set_lp_column($self->get_activity_registry->counter );
	$self->add_activities($new);
	return $new;
}
sub adhoc_solve {
	my $self = shift;
	my $milpcheck = shift;
	my $verbose = shift ||0;

	my @SortedAO = sort {$a->lp_column <=> $b->lp_column} $self->activities;
	my @SortedCO = sort {$a->lp_row <=> $b->lp_row} $self->constraints;
	my @SortedActivity;   
	my $ninteg = 0;
	foreach my $i(@SortedAO) {
		  push @SortedActivity , $i->id;   
		  $ninteg++ if ($i->get_attribute("integ") == 1);
	};
	my @SortedConstraint;   
	foreach my $i(@SortedCO) {push @SortedConstraint , $i->id  } ;
	$self->set_sorted_constraints(\@SortedConstraint);
	$self->set_sorted_activities(\@SortedActivity);
	
	my $milpinfo = {(
				NumCol => $self->get_activity_registry->counter,
				NumRow => $self->get_constraint_registry->counter,
				ZeroAct => 0,
				ZeroCon => 0,		
				IntegAct => $ninteg,
				Digs => 4
			)};
	$self->set_milpinfo($milpinfo);

	$self->blank_file;
	$self->prepare_special(1, "_", {});
	$self->write(1, "_",{},1) if $verbose;
	my $fn = $self->write(0, "_",{},1);
	
	$fn =~ s/.mtx$//;
	my $string = $milpcheck." ".$fn." -seq 4";
	$fn .= ".pri";
	unlink($fn);
	
	my $rm = qx($string);
	$self->{adhoc_solve_rm} = $rm;
	my $rv = 1;
	$rv = 0 if ($rm =~ /Unbounded/ or $rm =~ /Infeasible/) ;   
	eval {
		local $SIG{__WARN__}=sub{}; #temporarily suppress warnings
		$self->import_solution_vector($fn);
	};
	if ($@) {
	      croak "$@";
	      return 0;
	}
	else {
		#print "Solution has been imported.\n"
	  return $rv;
	}
}
sub adhoc_status_message {
    return $_[0]->{adhoc_solve_rm};
}
1;
